import io
import logging
import warnings
from abc import ABC, abstractmethod
from datetime import tzinfo
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    Generator,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Union,
)

import pytz
from pytz.exceptions import UnknownTimeZoneError

from clickhouse_connect import common
from clickhouse_connect.common import version
from clickhouse_connect.datatypes import dynamic as dynamic_module
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.driver import options, tzutil
from clickhouse_connect.driver.binding import quote_identifier
from clickhouse_connect.driver.common import (
    StreamContext,
    coerce_bool,
    coerce_int,
    dict_copy,
)
from clickhouse_connect.driver.constants import (
    CH_VERSION_WITH_PROTOCOL,
    PROTOCOL_VERSION_WITH_LOW_CARD,
)
from clickhouse_connect.driver.exceptions import (
    DataError,
    OperationalError,
    ProgrammingError,
)
from clickhouse_connect.driver.external import ExternalData
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.models import ColumnDef, SettingDef, SettingStatus
from clickhouse_connect.driver.options import (
    check_arrow,
    check_numpy,
    check_pandas,
    check_polars,
)
from clickhouse_connect.driver.query import (
    _APPLY_SERVER_TZ_TO_TZ_SOURCE,
    _TZ_MODE_TO_UTC_TZ_AWARE,
    _VALID_TZ_SOURCES,
    QueryContext,
    QueryResult,
    TzMode,
    TzSource,
    _resolve_tz_mode,
    _resolve_tz_source,
    arrow_buffer,
    to_arrow,
    to_arrow_batches,
)
from clickhouse_connect.driver.summary import QuerySummary

if TYPE_CHECKING:
    import numpy
    import pandas
    import pyarrow

io.DEFAULT_BUFFER_SIZE = 1024 * 256
logger = logging.getLogger(__name__)
arrow_str_setting = 'output_format_arrow_string_as_string'


def _strip_utc_timezone_from_arrow(table: "arrow.Table") -> "arrow.Table":
    """Strip UTC timezone from timestamp columns in Arrow table.

    This ensures naive datetimes are returned when the server timezone is UTC
    and utc_tz_aware is False (the default).

    Only UTC-equivalent timezones (UTC, Etc/UTC, GMT, etc.) are stripped.
    Non-UTC timezones carry important offset information and are always
    preserved regardless of utc_tz_aware setting.
    """
    new_fields = []
    needs_cast = False
    for field in table.schema:
        if options.arrow.types.is_timestamp(field.type) and tzutil.is_utc_timezone(field.type.tz):
            new_fields.append(options.arrow.field(field.name, options.arrow.timestamp(field.type.unit)))
            needs_cast = True
        else:
            new_fields.append(field)
    if needs_cast:
        return table.cast(options.arrow.schema(new_fields))
    return table


def _apply_arrow_tz_policy(table: "options.arrow.Table", tz_mode: str) -> "options.arrow.Table":
    """Apply the tz_mode policy to an Arrow table before conversion.

    Handles UTC stripping when tz_mode is "naive_utc" and warns when
    tz_mode is "schema" since that mode is not yet implemented for
    Arrow-based queries.
    """
    if tz_mode == "schema":
        logger.warning(
            'tz_mode="schema" is not yet supported for Arrow-based query methods. '
            "It would require a separate schema lookup since ClickHouse attaches the server "
            "timezone to all DateTime columns in Arrow format. Use query/query_df for "
            "schema-matching behavior or open an issue if you need Arrow support."
        )
    if tz_mode == "naive_utc":
        table = _strip_utc_timezone_from_arrow(table)
    return table


# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods,too-many-arguments,too-many-positional-arguments,too-many-instance-attributes
class Client(ABC):
    """
    Base ClickHouse Connect client
    """
    compression: str = None
    write_compression: str = None
    protocol_version = 0
    valid_transport_settings = set()
    optional_transport_settings = set()
    database = None
    max_error_message = 0
    _tz_source: TzSource = "auto"
    _apply_server_tz = False
    tz_mode: TzMode = "naive_utc"
    show_clickhouse_errors = True

    @property
    def tz_source(self) -> TzSource:
        return self._tz_source

    @tz_source.setter
    def tz_source(self, value: TzSource):
        if value not in _VALID_TZ_SOURCES:
            raise ProgrammingError(
                f'tz_source must be "auto", "server", or "local", got "{value}"'
            )
        self._tz_source = value
        if value == "auto":
            self._apply_server_tz = self._dst_safe
        else:
            self._apply_server_tz = value == "server"

    @property
    def apply_server_timezone(self) -> bool:
        """Deprecated: use tz_source instead."""
        warnings.warn(
            "apply_server_timezone is deprecated and will be removed in 1.0. "
            "Use tz_source instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._apply_server_tz

    @apply_server_timezone.setter
    def apply_server_timezone(self, value: Union[bool, str]):
        """Deprecated: use tz_source instead."""
        warnings.warn(
            "apply_server_timezone is deprecated and will be removed in 1.0. "
            "Use tz_source instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if value not in _APPLY_SERVER_TZ_TO_TZ_SOURCE:
            raise ProgrammingError(
                f"apply_server_timezone must be True, False, or 'always', got \"{value}\""
            )
        self.tz_source = _APPLY_SERVER_TZ_TO_TZ_SOURCE[value]

    @property
    def utc_tz_aware(self) -> Union[bool, Literal["schema"]]:
        """Deprecated: use tz_mode instead."""
        warnings.warn(
            "utc_tz_aware is deprecated and will be removed in 1.0. "
            "Use tz_mode instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return _TZ_MODE_TO_UTC_TZ_AWARE[self.tz_mode]

    def __init__(self,
                 database: str,
                 query_limit: int,
                 uri: str,
                 query_retries: int,
                 server_host_name: Optional[str],
                 tz_source: Optional[TzSource] = None,
                 tz_mode: Optional[TzMode] = None,
                 show_clickhouse_errors: Optional[bool] = None,
                 utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                 apply_server_timezone: Optional[Union[str, bool]] = None):
        """
        Shared initialization of ClickHouse Connect client
        :param database: database name
        :param query_limit: default LIMIT for queries
        :param uri: uri for error messages
        :param tz_source: Controls how the client determines the fallback timezone for DateTime columns without an
          explicit timezone. "auto" (default) auto-detects based on DST safety of server timezone. "server" always
          uses the server timezone. "local" always uses the local timezone.
        :param tz_mode: Controls timezone-aware behavior for UTC DateTime columns.  "naive_utc" (default) returns
          naive UTC timestamps.  "aware" forces timezone-aware UTC datetimes.  "schema" returns datetimes that
          match the server's column definition which means timezone-aware when the column defines a timezone and naive
          for bare DateTime columns.
        :param utc_tz_aware: Deprecated. Use tz_mode instead.
        :param apply_server_timezone: Deprecated. Use tz_source instead.
        """
        self.query_limit = coerce_int(query_limit)
        self.query_retries = coerce_int(query_retries)
        if database and not database == '__default__':
            self.database = database
        if show_clickhouse_errors is not None:
            self.show_clickhouse_errors = coerce_bool(show_clickhouse_errors)
        self.server_host_name = server_host_name
        self.uri = uri
        self.tz_mode = _resolve_tz_mode(tz_mode, utc_tz_aware)
        resolved_tz_source = _resolve_tz_source(tz_source, apply_server_timezone)
        self._tz_source = resolved_tz_source
        self._init_common_settings(resolved_tz_source)

    def _init_common_settings(self, tz_source: TzSource):
        self.server_tz, self._dst_safe = pytz.UTC, True
        self.server_version, server_tz = \
            tuple(self.command('SELECT version(), timezone()', use_database=False))
        try:
            server_tz = pytz.timezone(server_tz)
            server_tz, self._dst_safe = tzutil.normalize_timezone(server_tz)
            if tz_source == "auto":
                self._apply_server_tz = self._dst_safe
            else:
                self._apply_server_tz = tz_source == "server"
            self.server_tz = server_tz
        except UnknownTimeZoneError:
            logger.warning('Warning, server is using an unrecognized timezone %s, will use UTC default', server_tz)

        if not self._apply_server_tz and not tzutil.local_tz_dst_safe:
            logger.warning('local timezone %s may return unexpected times due to Daylight Savings Time/' +
                           'Summer Time differences', tzutil.local_tz.tzname(None))
        readonly = 'readonly'
        if not self.min_version('19.17'):
            readonly = common.get_setting('readonly')
        server_settings = self.query(f'SELECT name, value, {readonly} as readonly FROM system.settings LIMIT 10000')
        self.server_settings = {row['name']: SettingDef(**row) for row in server_settings.named_results()}

        if self.min_version(CH_VERSION_WITH_PROTOCOL) and common.get_setting('use_protocol_version'):
            #  Unfortunately we have to validate that the client protocol version is actually used by ClickHouse
            #  since the query parameter could be stripped off (in particular, by CHProxy)
            test_data = self.raw_query('SELECT 1 AS check', fmt='Native', settings={
                'client_protocol_version': PROTOCOL_VERSION_WITH_LOW_CARD
            })
            if test_data[8:16] == b'\x01\x01\x05check':
                self.protocol_version = PROTOCOL_VERSION_WITH_LOW_CARD
        if self._setting_status('date_time_input_format').is_writable:
            self.set_client_setting('date_time_input_format', 'best_effort')
        if self._setting_status('allow_experimental_json_type').is_set and \
                self._setting_status('cast_string_to_dynamic_use_inference').is_writable:
            self.set_client_setting('cast_string_to_dynamic_use_inference', '1')
        if self.min_version('24.8') and not self.min_version('24.10'):
            dynamic_module.json_serialization_format = 0

    def _validate_settings(self, settings: Optional[Dict[str, Any]]) -> Dict[str, str]:
        """
        This strips any ClickHouse settings that are not recognized or are read only.
        :param settings:  Dictionary of setting name and values
        :return: A filtered dictionary of settings with values rendered as strings
        """
        validated = {}
        invalid_action = common.get_setting('invalid_setting_action')
        for key, value in settings.items():
            str_value = self._validate_setting(key, value, invalid_action)
            if str_value is not None:
                validated[key] = value
        return validated

    def _validate_setting(self, key: str, value: Any, invalid_action: str) -> Optional[str]:
        str_value = str(value)
        if value is True:
            str_value = '1'
        elif value is False:
            str_value = '0'
        if key not in self.valid_transport_settings:
            setting_def = self.server_settings.get(key)
            current_setting = self.get_client_setting(key)
            # Skip if the requested value matches the server's value and either the setting
            # is readonly i.e. there's nothing to change or already explicitly stored on the client
            if setting_def and setting_def.value == str_value:
                if setting_def.readonly or (current_setting is not None and current_setting == setting_def.value):
                    return None
            if setting_def is None or setting_def.readonly:
                if key in self.optional_transport_settings:
                    return None
                if invalid_action == 'send':
                    logger.warning('Attempting to send unrecognized or readonly setting %s', key)
                elif invalid_action == 'drop':
                    logger.warning('Dropping unrecognized or readonly settings %s', key)
                    return None
                else:
                    raise ProgrammingError(f'Setting {key} is unknown or readonly') from None
        return str_value

    def _setting_status(self, key: str) -> SettingStatus:
        comp_setting = self.server_settings.get(key)
        if not comp_setting:
            return SettingStatus(False, False)
        return SettingStatus(comp_setting.value != '0', comp_setting.readonly != 1)

    def _prep_query(self, context: QueryContext):
        if context.is_select and not context.has_limit and self.query_limit:
            limit = f'\n LIMIT {self.query_limit}'
            if isinstance(context.query, bytes):
                return context.final_query + limit.encode()
            return context.final_query + limit
        return context.final_query

    def _check_tz_change(self, new_tz) -> Optional[tzinfo]:
        if new_tz:
            try:
                new_tzinfo = pytz.timezone(new_tz)
                if new_tzinfo != self.server_tz:
                    return new_tzinfo
            except UnknownTimeZoneError:
                logger.warning('Unrecognized timezone %s received from ClickHouse', new_tz)
        return None

    @abstractmethod
    def _query_with_context(self, context: QueryContext) -> QueryResult:
        pass

    @abstractmethod
    def set_client_setting(self, key: str, value: Any) -> None:
        """
        Set a clickhouse setting for the client after initialization.  If a setting is not recognized by ClickHouse,
        or the setting is identified as "read_only", this call will either throw a Programming exception or attempt
        to send the setting anyway based on the common setting 'invalid_setting_action'
        :param key: ClickHouse setting name
        :param value: ClickHouse setting value
        """

    @abstractmethod
    def get_client_setting(self, key: str) -> Optional[str]:
        """
        :param key: The setting key
        :return: The string value of the setting, if it exists, or None
        """

    @abstractmethod
    def set_access_token(self, access_token: str) -> None:
        """
        Set the ClickHouse access token for the client
        :param access_token: Access token string
        """

    # pylint: disable=unused-argument,too-many-locals
    def query(self,
              query: Optional[str] = None,
              parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
              settings: Optional[Dict[str, Any]] = None,
              query_formats: Optional[Dict[str, str]] = None,
              column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
              encoding: Optional[str] = None,
              use_none: Optional[bool] = None,
              column_oriented: Optional[bool] = None,
              use_numpy: Optional[bool] = None,
              max_str_len: Optional[int] = None,
              context: QueryContext = None,
              query_tz: Optional[Union[str, tzinfo]] = None,
              column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
              utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
              external_data: Optional[ExternalData] = None,
              transport_settings: Optional[Dict[str, str]] = None,
              tz_mode: Optional[TzMode] = None) -> QueryResult:
        """
        Main query method for SELECT, DESCRIBE and other SQL statements that return a result matrix.  For
        parameters, see the create_query_context method
        :return: QueryResult -- data and metadata from response
        """
        if query and query.lower().strip().startswith('select __connect_version__'):
            return QueryResult([[f'ClickHouse Connect v.{version()}  ⓒ ClickHouse Inc.']], None,
                               ('connect_version',), (get_from_name('String'),))
        kwargs = locals().copy()
        del kwargs['self']
        query_context = self.create_query_context(**kwargs)
        if query_context.is_command:
            response = self.command(query,
                                    parameters=query_context.parameters,
                                    settings=query_context.settings,
                                    external_data=query_context.external_data,
                                    transport_settings=query_context.transport_settings)
            if isinstance(response, QuerySummary):
                return response.as_query_result()
            return QueryResult([response] if isinstance(response, list) else [[response]])
        return self._query_with_context(query_context)

    def query_column_block_stream(self,
                                  query: Optional[str] = None,
                                  parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                                  settings: Optional[Dict[str, Any]] = None,
                                  query_formats: Optional[Dict[str, str]] = None,
                                  column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                                  encoding: Optional[str] = None,
                                  use_none: Optional[bool] = None,
                                  context: QueryContext = None,
                                  query_tz: Optional[Union[str, tzinfo]] = None,
                                  column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                                  utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                                  external_data: Optional[ExternalData] = None,
                                  transport_settings: Optional[Dict[str, str]] = None,
                                  tz_mode: Optional[TzMode] = None) -> StreamContext:
        """
        Variation of main query method that returns a stream of column oriented blocks. For
        parameters, see the create_query_context method.
        :return: StreamContext -- Iterable stream context that returns column oriented blocks
        """
        return self._context_query(locals(), use_numpy=False, streaming=True).column_block_stream

    def query_row_block_stream(self,
                               query: Optional[str] = None,
                               parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                               settings: Optional[Dict[str, Any]] = None,
                               query_formats: Optional[Dict[str, str]] = None,
                               column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                               encoding: Optional[str] = None,
                               use_none: Optional[bool] = None,
                               context: QueryContext = None,
                               query_tz: Optional[Union[str, tzinfo]] = None,
                               column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                               utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                               external_data: Optional[ExternalData] = None,
                               transport_settings: Optional[Dict[str, str]] = None,
                               tz_mode: Optional[TzMode] = None) -> StreamContext:
        """
        Variation of main query method that returns a stream of row oriented blocks. For
        parameters, see the create_query_context method.
        :return: StreamContext -- Iterable stream context that returns blocks of rows
        """
        return self._context_query(locals(), use_numpy=False, streaming=True).row_block_stream

    def query_rows_stream(self,
                          query: Optional[str] = None,
                          parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                          settings: Optional[Dict[str, Any]] = None,
                          query_formats: Optional[Dict[str, str]] = None,
                          column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                          encoding: Optional[str] = None,
                          use_none: Optional[bool] = None,
                          context: QueryContext = None,
                          query_tz: Optional[Union[str, tzinfo]] = None,
                          column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                          utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                          external_data: Optional[ExternalData] = None,
                          transport_settings: Optional[Dict[str, str]] = None,
                          tz_mode: Optional[TzMode] = None) -> StreamContext:
        """
        Variation of main query method that returns a stream of row oriented blocks. For
        parameters, see the create_query_context method.
        :return: StreamContext -- Iterable stream context that returns blocks of rows
        """
        return self._context_query(locals(), use_numpy=False, streaming=True).rows_stream

    @abstractmethod
    def raw_query(self, query: str,
                  parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                  settings: Optional[Dict[str, Any]] = None,
                  fmt: str = None,
                  use_database: bool = True,
                  external_data: Optional[ExternalData] = None,
                  transport_settings: Optional[Dict[str, str]] = None) -> bytes:
        """
        Query method that simply returns the raw ClickHouse format bytes
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param fmt: ClickHouse output format
        :param use_database: Send the database parameter to ClickHouse so the command will be executed in the client
         database context.
        :param external_data: External data to send with the query
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: bytes representing raw ClickHouse return value based on format
        """

    @abstractmethod
    def raw_stream(self, query: str,
                   parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                   settings: Optional[Dict[str, Any]] = None,
                   fmt: str = None,
                   use_database: bool = True,
                   external_data: Optional[ExternalData] = None,
                   transport_settings: Optional[Dict[str, str]] = None) -> io.IOBase:
        """
       Query method that returns the result as an io.IOBase iterator
       :param query: Query statement/format string
       :param parameters: Optional dictionary used to format the query
       :param settings: Optional dictionary of ClickHouse settings (key/string values)
       :param fmt: ClickHouse output format
       :param use_database  Send the database parameter to ClickHouse so the command will be executed in the client
        database context.
       :param external_data: External data to send with the query.
       :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
       :return: io.IOBase stream/iterator for the result
       """

    # pylint: disable=duplicate-code,unused-argument
    def query_np(self,
                 query: Optional[str] = None,
                 parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, str]] = None,
                 encoding: Optional[str] = None,
                 use_none: Optional[bool] = None,
                 max_str_len: Optional[int] = None,
                 context: QueryContext = None,
                 external_data: Optional[ExternalData] = None,
                 transport_settings: Optional[Dict[str, str]] = None) -> 'numpy.ndarray':
        """
        Query method that returns the results as a numpy array.  For parameter values, see the
        create_query_context method
        :return: Numpy array representing the result set
        """
        check_numpy()
        self._add_integration_tag("numpy")
        return self._context_query(locals(), use_numpy=True).np_result

    # pylint: disable=duplicate-code,too-many-arguments,unused-argument
    def query_np_stream(self,
                        query: Optional[str] = None,
                        parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                        settings: Optional[Dict[str, Any]] = None,
                        query_formats: Optional[Dict[str, str]] = None,
                        column_formats: Optional[Dict[str, str]] = None,
                        encoding: Optional[str] = None,
                        use_none: Optional[bool] = None,
                        max_str_len: Optional[int] = None,
                        context: QueryContext = None,
                        external_data: Optional[ExternalData] = None,
                        transport_settings: Optional[Dict[str, str]] = None) -> StreamContext:
        """
        Query method that returns the results as a stream of numpy arrays.  For parameter values, see the
        create_query_context method
        :return: Generator that yield a numpy array per block representing the result set
        """
        check_numpy()
        self._add_integration_tag("numpy")
        return self._context_query(locals(), use_numpy=True, streaming=True).np_stream

    # pylint: disable=duplicate-code,unused-argument
    def query_df(self,
                 query: Optional[str] = None,
                 parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, str]] = None,
                 encoding: Optional[str] = None,
                 use_none: Optional[bool] = None,
                 max_str_len: Optional[int] = None,
                 use_na_values: Optional[bool] = None,
                 query_tz: Optional[str] = None,
                 column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                 utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                 context: QueryContext = None,
                 external_data: Optional[ExternalData] = None,
                 use_extended_dtypes: Optional[bool] = None,
                 transport_settings: Optional[Dict[str, str]] = None,
                 tz_mode: Optional[TzMode] = None) -> 'pandas.DataFrame':
        """
        Query method that results the results as a pandas dataframe.  For parameter values, see the
        create_query_context method
        :return: Pandas dataframe representing the result set
        """
        check_pandas()
        self._add_integration_tag("pandas")
        return self._context_query(locals(), use_numpy=True, as_pandas=True).df_result

    # pylint: disable=duplicate-code,unused-argument
    def query_df_stream(self,
                        query: Optional[str] = None,
                        parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                        settings: Optional[Dict[str, Any]] = None,
                        query_formats: Optional[Dict[str, str]] = None,
                        column_formats: Optional[Dict[str, str]] = None,
                        encoding: Optional[str] = None,
                        use_none: Optional[bool] = None,
                        max_str_len: Optional[int] = None,
                        use_na_values: Optional[bool] = None,
                        query_tz: Optional[str] = None,
                        column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                        utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                        context: QueryContext = None,
                        external_data: Optional[ExternalData] = None,
                        use_extended_dtypes: Optional[bool] = None,
                        transport_settings: Optional[Dict[str, str]] = None,
                        tz_mode: Optional[TzMode] = None) -> StreamContext:
        """
        Query method that returns the results as a StreamContext.  For parameter values, see the
        create_query_context method
        :return: Generator that yields a Pandas dataframe per block representing the result set
        """
        check_pandas()
        self._add_integration_tag("pandas")
        return self._context_query(locals(), use_numpy=True,
                                   as_pandas=True,
                                   streaming=True).df_stream

    def create_query_context(self,
                             query: Optional[Union[str, bytes]] = None,
                             parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                             settings: Optional[Dict[str, Any]] = None,
                             query_formats: Optional[Dict[str, str]] = None,
                             column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                             encoding: Optional[str] = None,
                             use_none: Optional[bool] = None,
                             column_oriented: Optional[bool] = None,
                             use_numpy: Optional[bool] = False,
                             max_str_len: Optional[int] = 0,
                             context: Optional[QueryContext] = None,
                             query_tz: Optional[Union[str, tzinfo]] = None,
                             column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                             utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                             use_na_values: Optional[bool] = None,
                             streaming: bool = False,
                             as_pandas: bool = False,
                             external_data: Optional[ExternalData] = None,
                             use_extended_dtypes: Optional[bool] = None,
                             transport_settings: Optional[Dict[str, str]] = None,
                             tz_mode: Optional[TzMode] = None) -> QueryContext:
        """
        Creates or updates a reusable QueryContext object
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param query_formats: See QueryContext __init__ docstring
        :param column_formats: See QueryContext __init__ docstring
        :param encoding: See QueryContext __init__ docstring
        :param use_none: Use None for ClickHouse NULL instead of default values.  Note that using None in Numpy
          arrays will force the numpy array dtype to 'object', which is often inefficient.  This effect also
          will impact the performance of Pandas dataframes.
        :param column_oriented: Deprecated. Controls orientation of the QueryResult result_set property
        :param use_numpy: Return QueryResult columns as one-dimensional numpy arrays
        :param max_str_len: Limit returned ClickHouse String values to this length, which allows a Numpy
          structured array even with ClickHouse variable length String columns.  If 0, Numpy arrays for
          String columns will always be object arrays
        :param context: An existing QueryContext to be updated with any provided parameter values
        :param query_tz: Either a string or a pytz tzinfo object.  (Strings will be converted to tzinfo objects).
          Values for any DateTime or DateTime64 column in the query will be converted to Python datetime.datetime
          objects with the selected timezone.
        :param column_tzs: A dictionary of column names to tzinfo objects (or strings that will be converted to
          tzinfo objects).  The timezone will be applied to datetime objects returned in the query
        :param tz_mode: Override the client default for handling UTC results.  "aware" forces timezone-aware
          UTC datetimes, "naive_utc" returns naive UTC datetimes, and "schema" returns datetimes matching the
          server's column definition.
        :param utc_tz_aware: Deprecated. Use tz_mode instead.
        :param use_na_values: Deprecated alias for use_advanced_dtypes
        :param as_pandas Return the result columns as pandas.Series objects
        :param streaming Marker used to correctly configure streaming queries
        :param external_data ClickHouse "external data" to send with query
        :param use_extended_dtypes:  Only relevant to Pandas Dataframe queries.  Use Pandas "missing types", such as
          pandas.NA and pandas.NaT for ClickHouse NULL values, as well as extended Pandas dtypes such as IntegerArray
          and StringArray.  Defaulted to True for query_df methods
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: Reusable QueryContext
        """
        if tz_mode is not None or utc_tz_aware is not None:
            resolved_tz_mode = _resolve_tz_mode(tz_mode, utc_tz_aware)
        else:
            resolved_tz_mode = self.tz_mode
        if context:
            return context.updated_copy(query=query,
                                        parameters=parameters,
                                        settings=settings,
                                        query_formats=query_formats,
                                        column_formats=column_formats,
                                        encoding=encoding,
                                        server_tz=self.server_tz,
                                        use_none=use_none,
                                        column_oriented=column_oriented,
                                        use_numpy=use_numpy,
                                        max_str_len=max_str_len,
                                        query_tz=query_tz,
                                        column_tzs=column_tzs,
                                        tz_mode=resolved_tz_mode,
                                        as_pandas=as_pandas,
                                        use_extended_dtypes=use_extended_dtypes,
                                        streaming=streaming,
                                        external_data=external_data,
                                        transport_settings=transport_settings)
        if use_numpy and max_str_len is None:
            max_str_len = 0
        if use_extended_dtypes is None:
            use_extended_dtypes = use_na_values
        if as_pandas and use_extended_dtypes is None:
            use_extended_dtypes = True
        return QueryContext(query=query,
                            parameters=parameters,
                            settings=settings,
                            query_formats=query_formats,
                            column_formats=column_formats,
                            encoding=encoding,
                            server_tz=self.server_tz,
                            use_none=use_none,
                            column_oriented=column_oriented,
                            use_numpy=use_numpy,
                            max_str_len=max_str_len,
                            query_tz=query_tz,
                            column_tzs=column_tzs,
                            tz_mode=resolved_tz_mode,
                            use_extended_dtypes=use_extended_dtypes,
                            as_pandas=as_pandas,
                            streaming=streaming,
                            apply_server_tz=self._apply_server_tz,
                            external_data=external_data,
                            transport_settings=transport_settings)

    def query_arrow(self,
                    query: str,
                    parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                    settings: Optional[Dict[str, Any]] = None,
                    use_strings: Optional[bool] = None,
                    external_data: Optional[ExternalData] = None,
                    transport_settings: Optional[Dict[str, str]] = None) -> 'pyarrow.Table':
        """
        Query method using the ClickHouse Arrow format to return a PyArrow table
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_strings: Convert ClickHouse String type to Arrow string type (instead of binary)
        :param external_data: ClickHouse "external data" to send with query
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: PyArrow.Table
        """
        check_arrow()
        self._add_integration_tag("arrow")
        settings = self._update_arrow_settings(settings, use_strings)
        return to_arrow(self.raw_query(query,
                                       parameters,
                                       settings,
                                       fmt='Arrow',
                                       external_data=external_data,
                                       transport_settings=transport_settings))

    def query_arrow_stream(self,
                           query: str,
                           parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                           settings: Optional[Dict[str, Any]] = None,
                           use_strings: Optional[bool] = None,
                           external_data: Optional[ExternalData] = None,
                           transport_settings: Optional[Dict[str, str]] = None) -> StreamContext:
        """
        Query method that returns the results as a stream of Arrow tables
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_strings: Convert ClickHouse String type to Arrow string type (instead of binary)
        :param external_data: ClickHouse "external data" to send with query
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: Generator that yields a PyArrow.Table for per block representing the result set
        """
        check_arrow()
        self._add_integration_tag("arrow")
        settings = self._update_arrow_settings(settings, use_strings)
        return to_arrow_batches(self.raw_stream(query,
                                                parameters,
                                                settings,
                                                fmt='ArrowStream',
                                                external_data=external_data,
                                                transport_settings=transport_settings))

    def query_df_arrow(self,
                       query: str,
                       parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                       settings: Optional[Dict[str, Any]] = None,
                       use_strings: Optional[bool] = None,
                       external_data: Optional[ExternalData] = None,
                       transport_settings: Optional[Dict[str, str]] = None,
                       dataframe_library: str = "pandas"
                       ) -> Union["pd.DataFrame", "pl.DataFrame"]:
        """
        Query method using the ClickHouse Arrow format to return a DataFrame
        with PyArrow dtype backend. This provides better performance and memory efficiency
        compared to the standard query_df method, though fewer output formatting options.

        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_strings: Convert ClickHouse String type to Arrow string type (instead of binary)
        :param external_data: ClickHouse "external data" to send with query
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :param dataframe_library: Library to use for DataFrame creation ("pandas" or "polars")
        :return: DataFrame (pandas or polars based on dataframe_library parameter)
        """
        check_arrow()

        if dataframe_library == "pandas":
            check_pandas()
            self._add_integration_tag("pandas")
            if not options.IS_PANDAS_2:
                raise ProgrammingError("PyArrow-backed dtypes are only supported when using pandas 2.x.")

            def converter(table: options.arrow.Table) -> options.pd.DataFrame:
                table = _apply_arrow_tz_policy(table, self.tz_mode)
                return table.to_pandas(types_mapper=options.pd.ArrowDtype, safe=False)

        elif dataframe_library == "polars":
            check_polars()
            self._add_integration_tag("polars")

            def converter(table: options.arrow.Table) -> options.pl.DataFrame:
                table = _apply_arrow_tz_policy(table, self.tz_mode)
                return options.pl.from_arrow(table)

        else:
            raise ValueError(f"dataframe_library must be 'pandas' or 'polars', got '{dataframe_library}'")

        arrow_table = self.query_arrow(
            query=query,
            parameters=parameters,
            settings=settings,
            use_strings=use_strings,
            external_data=external_data,
            transport_settings=transport_settings,
        )

        return converter(arrow_table)

    def query_df_arrow_stream(self,
                             query: str,
                             parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                             settings: Optional[Dict[str, Any]] = None,
                             use_strings: Optional[bool] = None,
                             external_data: Optional[ExternalData] = None,
                             transport_settings: Optional[Dict[str, str]] = None,
                             dataframe_library: str = "pandas") -> StreamContext:
        """
        Query method that returns the results as a stream of DataFrames with PyArrow dtype backend.
        Each DataFrame represents a block from the ClickHouse response.

        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_strings: Convert ClickHouse String type to Arrow string type (instead of binary)
        :param external_data: ClickHouse "external data" to send with query
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :param dataframe_library: Library to use for DataFrame creation ("pandas" or "polars")
        :return: StreamContext that yields DataFrames (pandas or polars based on dataframe_library parameter)
        """
        check_arrow()
        if dataframe_library == "pandas":
            check_pandas()
            self._add_integration_tag("pandas")
            if not options.IS_PANDAS_2:
                raise ProgrammingError("PyArrow-backed dtypes are only supported when using pandas 2.x.")

            def converter(table: "options.arrow.Table") -> "options.pd.DataFrame":
                table = _apply_arrow_tz_policy(table, self.tz_mode)
                return table.to_pandas(types_mapper=options.pd.ArrowDtype, safe=False)
        elif dataframe_library == "polars":
            check_polars()
            self._add_integration_tag("polars")

            def converter(table: options.arrow.Table) -> options.pl.DataFrame:
                table = _apply_arrow_tz_policy(table, self.tz_mode)
                return options.pl.from_arrow(table)
        else:
            raise ValueError(f"dataframe_library must be 'pandas' or 'polars', got '{dataframe_library}'")
        settings = self._update_arrow_settings(settings, use_strings)
        raw_stream = self.raw_stream(
            query, parameters, settings, fmt="ArrowStream", external_data=external_data, transport_settings=transport_settings
        )
        reader = options.arrow.ipc.open_stream(raw_stream)

        def df_generator():
            for batch in reader:
                yield converter(batch)

        return StreamContext(raw_stream, df_generator())

    def _update_arrow_settings(self,
                               settings: Optional[Dict[str, Any]],
                               use_strings: Optional[bool]) -> Dict[str, Any]:
        settings = dict_copy(settings)
        if self.database:
            settings['database'] = self.database
        str_status = self._setting_status(arrow_str_setting)
        if use_strings is None:
            if str_status.is_writable and not str_status.is_set:
                settings[arrow_str_setting] = '1'  # Default to returning strings if possible
        elif use_strings != str_status.is_set:
            if not str_status.is_writable:
                raise OperationalError(f'Cannot change readonly {arrow_str_setting} to {use_strings}')
            settings[arrow_str_setting] = '1' if use_strings else '0'
        return settings

    @abstractmethod
    def command(self,
                cmd: str,
                parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                data: Union[str, bytes] = None,
                settings: Dict[str, Any] = None,
                use_database: bool = True,
                external_data: Optional[ExternalData] = None,
                transport_settings: Optional[Dict[str, str]] = None) -> Union[str, int, Sequence[str], QuerySummary]:
        """
        Client method that returns a single value instead of a result set
        :param cmd: ClickHouse query/command as a python format string
        :param parameters: Optional dictionary of key/values pairs to be formatted
        :param data: Optional 'data' for the command (for INSERT INTO in particular)
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_database: Send the database parameter to ClickHouse so the command will be executed in the client
         database context. Otherwise, no database will be specified with the command.  This is useful for determining
         the default user database
        :param external_data: ClickHouse "external data" to send with command/query
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: Decoded response from ClickHouse as either a string, int, or sequence of strings, or QuerySummary
        if no data returned
        """

    @abstractmethod
    def ping(self) -> bool:
        """
        Validate the connection, does not throw an Exception (see debug logs)
        :return: ClickHouse server is up and reachable
        """

    def insert(self,
               table: Optional[str] = None,
               data: Sequence[Sequence[Any]] = None,
               column_names: Union[str, Iterable[str]] = '*',
               database: Optional[str] = None,
               column_types: Sequence[ClickHouseType] = None,
               column_type_names: Sequence[str] = None,
               column_oriented: bool = False,
               settings: Optional[Dict[str, Any]] = None,
               context: InsertContext = None,
               transport_settings: Optional[Dict[str, str]] = None) -> QuerySummary:
        """
        Method to insert multiple rows/data matrix of native Python objects.  If context is specified arguments
        other than data are ignored
        :param table: Target table
        :param data: Sequence of sequences of Python data
        :param column_names: Ordered list of column names or '*' if column types should be retrieved from the
            ClickHouse table definition
        :param database: Target database -- will use client default database if not specified.
        :param column_types: ClickHouse column types.  If set then column data does not need to be retrieved from
            the server
        :param column_type_names: ClickHouse column type names.  If set then column data does not need to be
            retrieved from the server
        :param column_oriented: If true the data is already "pivoted" in column form
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param context: Optional reusable insert context to allow repeated inserts into the same table with
            different data batches
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: QuerySummary with summary information, throws exception if insert fails
        """
        if (context is None or context.empty) and data is None:
            raise ProgrammingError('No data specified for insert') from None
        if context is None:
            context = self.create_insert_context(table,
                                                 column_names,
                                                 database,
                                                 column_types,
                                                 column_type_names,
                                                 column_oriented,
                                                 settings,
                                                 transport_settings=transport_settings)
        if data is not None:
            if not context.empty:
                raise ProgrammingError('Attempting to insert new data with non-empty insert context') from None
            context.data = data
        return self.data_insert(context)

    def insert_df(self, table: str = None,
                  df=None,
                  database: Optional[str] = None,
                  settings: Optional[Dict] = None,
                  column_names: Optional[Sequence[str]] = None,
                  column_types: Sequence[ClickHouseType] = None,
                  column_type_names: Sequence[str] = None,
                  context: InsertContext = None,
                  transport_settings: Optional[Dict[str, str]] = None) -> QuerySummary:
        """
        Insert a pandas DataFrame into ClickHouse.  If context is specified arguments other than df are ignored
        :param table: ClickHouse table
        :param df: two-dimensional pandas dataframe
        :param database: Optional ClickHouse database
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param column_names: An optional list of ClickHouse column names.  If not set, the DataFrame column names
           will be used
        :param column_types: ClickHouse column types.  If set then column data does not need to be retrieved from
            the server
        :param column_type_names: ClickHouse column type names.  If set then column data does not need to be
            retrieved from the server
        :param context: Optional reusable insert context to allow repeated inserts into the same table with
            different data batches
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: QuerySummary with summary information, throws exception if insert fails
        """
        check_pandas()
        self._add_integration_tag("pandas")
        if context is None:
            if column_names is None:
                column_names = df.columns
            elif len(column_names) != len(df.columns):
                raise ProgrammingError('DataFrame column count does not match insert_columns') from None
        return self.insert(table,
                           df,
                           column_names,
                           database,
                           column_types=column_types,
                           column_type_names=column_type_names,
                           settings=settings,
                           transport_settings=transport_settings,
                           context=context)

    def insert_arrow(self, table: str,
                     arrow_table,
                     database: str = None,
                     settings: Optional[Dict] = None,
                     transport_settings: Optional[Dict[str, str]] = None) -> QuerySummary:
        """
        Insert a PyArrow table DataFrame into ClickHouse using raw Arrow format
        :param table: ClickHouse table
        :param arrow_table: PyArrow Table object
        :param database: Optional ClickHouse database
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        """
        check_arrow()
        self._add_integration_tag("arrow")
        full_table = table if '.' in table or not database else f'{database}.{table}'
        compression = self.write_compression if self.write_compression in ('zstd', 'lz4') else None
        column_names, insert_block = arrow_buffer(arrow_table, compression)
        return self.raw_insert(full_table, column_names, insert_block, settings, 'Arrow', transport_settings)

    def insert_df_arrow(self,
                        table: str,
                        df: Union["pd.DataFrame", "pl.DataFrame"],
                        database: Optional[str] = None,
                        settings: Optional[Dict] = None,
                        transport_settings: Optional[Dict[str, str]] = None) -> QuerySummary:
        """
        Insert a pandas DataFrame with PyArrow backend or a polars DataFrame into ClickHouse using Arrow format.
        This method is optimized for DataFrames that already use Arrow format, providing
        better performance than the standard insert_df method.
        
        Validation is performed and an exception will be raised if this requirement is not met.
        Polars DataFrames are natively Arrow-based and don't require additional validation.
        
        :param table: ClickHouse table name
        :param df: Pandas DataFrame with PyArrow dtype backend or Polars DataFrame
        :param database: Optional ClickHouse database name
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: QuerySummary with summary information, throws exception if insert fails
        """
        check_arrow()

        if options.pd is not None and isinstance(df, options.pd.DataFrame):
            df_lib = "pandas"
        elif options.pl is not None and isinstance(df, options.pl.DataFrame):
            df_lib = "polars"
        else:
            if options.pd is None and options.pl is None:
                raise ImportError("A DataFrame library (pandas or polars) must be installed to use insert_df_arrow.")
            raise TypeError(f"df must be either a pandas DataFrame or polars DataFrame, got {type(df).__name__}")

        if df_lib == "pandas":
            if not options.IS_PANDAS_2:
                raise ProgrammingError("PyArrow-backed dtypes are only supported when using pandas 2.x.")

            non_arrow_cols = [col for col, dtype in df.dtypes.items() if not isinstance(dtype, options.pd.ArrowDtype)]
            if non_arrow_cols:
                raise ProgrammingError(
                    f"insert_df_arrow requires all columns to use PyArrow dtypes. Non-Arrow columns found: [{', '.join(non_arrow_cols)}]. "
                )
            try:
                arrow_table = options.arrow.Table.from_pandas(df, preserve_index=False)
            except Exception as e:
                raise DataError(f"Failed to convert pandas DataFrame to Arrow table: {e}") from e
        else:
            try:
                arrow_table = df.to_arrow()
            except Exception as e:
                raise DataError(f"Failed to convert polars DataFrame to Arrow table: {e}") from e

        self._add_integration_tag(df_lib)
        return self.insert_arrow(
            table=table,
            arrow_table=arrow_table,
            database=database,
            settings=settings,
            transport_settings=transport_settings,
        )

    def create_insert_context(self,
                              table: str,
                              column_names: Optional[Union[str, Sequence[str]]] = None,
                              database: Optional[str] = None,
                              column_types: Sequence[ClickHouseType] = None,
                              column_type_names: Sequence[str] = None,
                              column_oriented: bool = False,
                              settings: Optional[Dict[str, Any]] = None,
                              data: Optional[Sequence[Sequence[Any]]] = None,
                              transport_settings: Optional[Dict[str, str]] = None) -> InsertContext:
        """
        Builds a reusable insert context to hold state for a duration of an insert
        :param table: Target table
        :param database: Target database.  If not set, uses the client default database
        :param column_names: Optional ordered list of column names.  If not set, all columns ('*') will be assumed
          in the order specified by the table definition
        :param database: Target database -- will use client default database if not specified
        :param column_types: ClickHouse column types.  Optional  Sequence of ClickHouseType objects.  If neither column
           types nor column type names are set, actual column types will be retrieved from the server.
        :param column_type_names: ClickHouse column type names.  Specified column types by name string
        :param column_oriented: If true the data is already "pivoted" in column form
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param data: Initial dataset for insert
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        :return: Reusable insert context
        """
        full_table = table
        if '.' not in table:
            if database:
                full_table = f'{quote_identifier(database)}.{quote_identifier(table)}'
            else:
                full_table = quote_identifier(table)
        column_defs = []
        if column_types is None and column_type_names is None:
            describe_result = self.query(f'DESCRIBE TABLE {full_table}', settings=settings)
            column_defs = [ColumnDef(**row) for row in describe_result.named_results()
                           if row['default_type'] not in ('ALIAS', 'MATERIALIZED')]
        if column_names is None or isinstance(column_names, str) and column_names == '*':
            column_names = [cd.name for cd in column_defs]
            column_types = [cd.ch_type for cd in column_defs]
        elif isinstance(column_names, str):
            column_names = [column_names]
        if len(column_names) == 0:
            raise ValueError('Column names must be specified for insert')
        if not column_types:
            if column_type_names:
                column_types = [get_from_name(name) for name in column_type_names]
            else:
                column_map = {d.name: d for d in column_defs}
                try:
                    column_types = [column_map[name].ch_type for name in column_names]
                except KeyError as ex:
                    raise ProgrammingError(f'Unrecognized column {ex} in table {table}') from None
        if len(column_names) != len(column_types):
            raise ProgrammingError('Column names do not match column types') from None
        return InsertContext(full_table,
                             column_names,
                             column_types,
                             column_oriented=column_oriented,
                             settings=settings,
                             transport_settings=transport_settings,
                             data=data)

    def min_version(self, version_str: str) -> bool:
        """
        Determine whether the connected server is at least the submitted version
        For Altinity Stable versions like 22.8.15.25.altinitystable
        the last condition in the first list comprehension expression is added
        :param version_str: A version string consisting of up to 4 integers delimited by dots
        :return: True if version_str is greater than the server_version, False if less than
        """
        try:
            server_parts = [int(x) for x in self.server_version.split('.') if x.isnumeric()]
            server_parts.extend([0] * (4 - len(server_parts)))
            version_parts = [int(x) for x in version_str.split('.')]
            version_parts.extend([0] * (4 - len(version_parts)))
        except ValueError:
            logger.warning('Server %s or requested version %s does not match format of numbers separated by dots',
                           self.server_version, version_str)
            return False
        for x, y in zip(server_parts, version_parts):
            if x > y:
                return True
            if x < y:
                return False
        return True

    # pylint: disable=no-self-use
    def _add_integration_tag(self, name: str) -> None:
        """Transport hook to surface 3rd party lib integration info (default: no-op)."""
        return

    @abstractmethod
    def data_insert(self, context: InsertContext) -> QuerySummary:
        """
        Subclass implementation of the data insert
        :context: InsertContext parameter object
        :return: No return, throws an exception if the insert fails
        """

    @abstractmethod
    def raw_insert(self, table: str,
                   column_names: Optional[Sequence[str]] = None,
                   insert_block: Union[str, bytes, Generator[bytes, None, None], BinaryIO] = None,
                   settings: Optional[Dict] = None,
                   fmt: Optional[str] = None,
                   compression: Optional[str] = None,
                   transport_settings: Optional[Dict[str, str]] = None) -> QuerySummary:
        """
        Insert data already formatted in a bytes object
        :param table: Table name (whether qualified with the database name or not)
        :param column_names: Sequence of column names
        :param insert_block: Binary or string data already in a recognized ClickHouse format
        :param settings:  Optional dictionary of ClickHouse settings (key/string values)
        :param fmt: Valid clickhouse format
        :param compression:  Recognized ClickHouse `Accept-Encoding` header compression value
        :param transport_settings: Optional dictionary of transport level settings (HTTP headers, etc.)
        """

    @abstractmethod
    def close(self) -> None:
        """
        Subclass implementation to close the connection to the server/deallocate the client
        """

    @abstractmethod
    def close_connections(self) -> None:
        """
        Subclass implementation to disconnect all "re-used" client connections
        """

    def _context_query(self, lcls: dict, **overrides):
        kwargs = lcls.copy()
        kwargs.pop('self')
        kwargs.update(overrides)
        return self._query_with_context((self.create_query_context(**kwargs)))

    def __enter__(self) -> 'Client':
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()



from clickhouse_connect.datatypes import registry, dynamic, geometric
from clickhouse_connect.datatypes.base import TypeDef
from clickhouse_connect.datatypes.container import Map

dynamic.STRING_DATA_TYPE = registry.get_from_name('String')

# Build a private Map(String, String) for JSON shared data decoding.
# We must NOT reuse the cached registry instance because we replace
# value_type with SharedDataString (reads raw bytes, encoding=None).
# Mutating the cached instance would break all normal Map(String, String) columns.
_shared_map = Map(TypeDef((), (), ('String', 'String')))
_shared_map.value_type = dynamic.SharedDataString(dynamic.STRING_DATA_TYPE.type_def)
dynamic.SHARED_DATA_TYPE = _shared_map

dynamic.SHARED_VARIANT_TYPE = dynamic.SharedVariant(dynamic.STRING_DATA_TYPE.type_def)

point = 'Tuple(Float64, Float64)'
ring = f'Array({point})'
polygon = f'Array({ring})'
multi_polygon = f'Array({polygon})'

geometric.POINT_DATA_TYPE = registry.get_from_name(point)
geometric.RING_DATA_TYPE = registry.get_from_name(ring)
geometric.POLYGON_DATA_TYPE = registry.get_from_name(polygon)
geometric.MULTI_POLYGON_DATA_TYPE = registry.get_from_name(multi_polygon)



import sys
from typing import Iterable, Any, Optional

import cython

from cpython cimport Py_INCREF, array
import array
from cpython.unicode cimport PyUnicode_Decode
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython.buffer cimport PyObject_GetBuffer, PyBuffer_Release, PyBUF_ANY_CONTIGUOUS, PyBUF_SIMPLE
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from libc.string cimport memcpy

from clickhouse_connect.driver.exceptions import StreamCompleteException

cdef union ull_wrapper:
    char* source
    unsigned long long int_value

cdef char * errors = 'strict'
cdef char * utf8 = 'utf8'
cdef dict array_templates = {}
cdef bint must_swap = sys.byteorder == 'big'
cdef array.array swapper = array.array('Q', [0])

for c in 'bBuhHiIlLqQfd':
    array_templates[c] = array.array(c, [])


cdef class ResponseBuffer:
    def __init__(self, source):
        self.slice_sz = 4096
        self.buf_loc = 0
        self.buf_sz = 0
        self.source = source
        self.gen = source.gen
        self.buffer = NULL
        self.slice = <char*>PyMem_Malloc(self.slice_sz)
        self._exception_tag = getattr(source, "exception_tag", None)
        self.open_marker = None
        self.close_marker = None
        self.carryover = b""
        self.exception_buf = None
        self.last_message_data = None
        self.current_chunk = None
        if self._exception_tag:
            tag_bytes = self._exception_tag.encode()
            self.open_marker = b"__exception__" + tag_bytes
            self.close_marker = tag_bytes + b"__exception__"

    cdef void _check_for_exception(self, object chunk) except *:
        cdef object search_data
        cdef Py_ssize_t marker_pos
        cdef Py_ssize_t carry_size
        if not self._exception_tag:
            return
        if self.exception_buf is not None:
            self.exception_buf += chunk
            if self.close_marker in self.exception_buf:
                self.last_message_data = bytes(self.exception_buf)
                raise StreamCompleteException
            return
        search_data = self.carryover + chunk
        marker_pos = search_data.find(self.open_marker)
        if marker_pos != -1:
            self.exception_buf = bytearray(search_data[marker_pos:])
            if self.close_marker in self.exception_buf:
                self.last_message_data = bytes(self.exception_buf)
                raise StreamCompleteException
            return
        carry_size = len(self.open_marker) - 1
        if len(search_data) >= carry_size:
            self.carryover = search_data[-carry_size:]
        else:
            self.carryover = search_data

    # Note that return char * return from this method is only good until the next call to _read_bytes_c or
    # _read_byte_load, since it points into self.buffer which can be replaced with the next chunk from the stream
    # Accordingly, that memory MUST be copied/processed into another buffer/PyObject immediately
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef char * read_bytes_c(self, unsigned long long sz) except NULL:
        cdef unsigned long long x, e, tail, cur_len, temp
        cdef char* ptr
        e = self.buf_sz

        if self.buf_loc + sz <= e:
            # We still have "sz" unread bytes available in the buffer, return the currently loc and advance it
            temp = self.buf_loc
            self.buf_loc += sz
            return self.buffer + temp

        # We need more data than is currently in the buffer, copy what's left into the temporary slice,
        # get a new buffer, and append what we need from the new buffer into that slice
        cur_len = e - self.buf_loc
        temp = self.slice_sz #
        while temp < sz * 2:
            temp <<= 1
        if temp > self.slice_sz:
            PyMem_Free(self.slice)
            self.slice = <char*>PyMem_Malloc(temp)
            self.slice_sz = temp
        if cur_len > 0:
            memcpy(self.slice, self.buffer + self.buf_loc, cur_len)
        self.buf_loc = 0
        self.buf_sz = 0

        # Loop until we've read enough chunks to fill the requested size
        while cur_len < sz:
            chunk = next(self.gen, None)
            if not chunk:
                raise StreamCompleteException
            self._check_for_exception(chunk)
            x = len(chunk)
            ptr = <char *> chunk
            if cur_len + x <= sz:
                # We need this whole chunk for the requested size, copy it into the temporary slice and get the next one
                memcpy(self.slice + cur_len, ptr, x)
                cur_len += x
            else:
                # We need just the beginning of this chunk to finish the temporary, copy that and set
                # the pointer into our stored buffer to the first unread data
                tail = sz - cur_len
                memcpy(self.slice + cur_len, ptr, tail)
                PyBuffer_Release(&self.buff_source)
                PyObject_GetBuffer(chunk, &self.buff_source, PyBUF_SIMPLE | PyBUF_ANY_CONTIGUOUS)
                self.buffer = <char *> self.buff_source.buf
                self.buf_sz = x
                self.buf_loc = tail
                self.current_chunk = chunk
                cur_len += tail
        return self.slice

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef inline unsigned char _read_byte_load(self) except ?255:
        self.buf_loc = 0
        self.buf_sz = 0
        chunk = next(self.gen, None)
        if not chunk:
            raise StreamCompleteException
        self._check_for_exception(chunk)
        x = len(chunk)
        self.current_chunk = chunk
        if x > 1:
            PyBuffer_Release(&self.buff_source)
            PyObject_GetBuffer(chunk, &self.buff_source, PyBUF_SIMPLE | PyBUF_ANY_CONTIGUOUS)
            self.buffer = <char *> self.buff_source.buf
            self.buf_loc = 1
            self.buf_sz = x
        return <unsigned char>chunk[0]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef inline object _read_str_col(self, unsigned long long num_rows, char * encoding):
        cdef object column = PyTuple_New(num_rows), v
        cdef unsigned long long x = 0, sz, shift
        cdef unsigned char b
        cdef char* buf
        while x < num_rows:
            sz = 0
            shift = 0
            while 1:
                if self.buf_loc < self.buf_sz:
                    b = self.buffer[self.buf_loc]
                    self.buf_loc += 1
                else:
                    b = self._read_byte_load()
                sz += ((b & 0x7f) << shift)
                if (b & 0x80) == 0:
                    break
                shift += 7
            buf = self.read_bytes_c(sz)
            if encoding:
                try:
                    v = PyUnicode_Decode(buf, sz, encoding, errors)
                except UnicodeDecodeError:
                    v = PyBytes_FromStringAndSize(buf, sz).hex()
            else:
                v = PyBytes_FromStringAndSize(buf, sz)
            PyTuple_SET_ITEM(column, x, v)
            Py_INCREF(v)
            x += 1
        return column

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef inline object _read_nullable_str_col(self, unsigned long long num_rows, char * encoding, object null_obj):
        cdef object column = PyTuple_New(num_rows), v
        cdef unsigned long long x = 0, sz, shift
        cdef unsigned char b
        cdef char * buf
        cdef char * null_map = <char *> PyMem_Malloc(<size_t> num_rows)
        memcpy(<void *> null_map, <void *> self.read_bytes_c(num_rows), num_rows)
        for x in range(num_rows):
            if self.buf_loc < self.buf_sz:
                b = self.buffer[self.buf_loc]
                self.buf_loc += 1
            else:
                b = self._read_byte_load()
            shift = 0
            sz = b & 0x7f
            while b & 0x80:
                shift += 7
                if self.buf_loc < self.buf_sz:
                    b = self.buffer[self.buf_loc]
                    self.buf_loc += 1
                else:
                    b = self._read_byte_load()
                sz += ((b & 0x7f) << shift)
            buf = self.read_bytes_c(sz)
            if null_map[x]:
                v = null_obj
            elif encoding:
                try:
                    v = PyUnicode_Decode(buf, sz, encoding, errors)
                except UnicodeDecodeError:
                    v = PyBytes_FromStringAndSize(buf, sz).hex()
            else:
                v = PyBytes_FromStringAndSize(buf, sz)
            PyTuple_SET_ITEM(column, x, v)
            Py_INCREF(v)
        PyMem_Free(<void *> null_map)
        return column

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_byte(self) -> int:
        if self.buf_loc < self.buf_sz:
            b = self.buffer[self.buf_loc]
            self.buf_loc += 1
            return b
        b = self._read_byte_load()
        return b

    def read_leb128_str(self) -> str:
        cdef unsigned long long sz = self.read_leb128()
        cdef char * b = self.read_bytes_c(sz)
        return PyUnicode_Decode(b, sz, utf8, errors)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_leb128(self) -> int:
        cdef:
            unsigned long long sz = 0, shift = 0
            unsigned char b
        while 1:
            if self.buf_loc < self.buf_sz:
                b = self.buffer[self.buf_loc]
                self.buf_loc += 1
            else:
                b = self._read_byte_load()
            sz += ((b & 0x7f) << shift)
            if (b & 0x80) == 0:
                return sz
            shift += 7

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_uint64(self) -> int:
        cdef ull_wrapper* x
        cdef char* b = self.read_bytes_c(8)
        if must_swap:
            memcpy(swapper.data.as_voidptr, b, 8)
            swapper.byteswap()
            return swapper[0]
        x = <ull_wrapper *> b
        return x.int_value

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_bytes(self, unsigned long long sz) -> bytes:
        cdef char* b = self.read_bytes_c(sz)
        return b[:sz]

    def read_str_col(self,
                     unsigned long long num_rows,
                     encoding: Optional[str],
                     nullable: bool = False,
                     null_object: Any = None) -> Iterable[str]:
        cdef char * enc = NULL
        if encoding:
            pyenc = encoding.encode()
            enc = pyenc
        if nullable:
            return self._read_nullable_str_col(num_rows, enc, null_object)
        return self._read_str_col(num_rows, enc)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_array(self, t: str, unsigned long long num_rows) -> Iterable[Any]:
        cdef array.array template = array_templates[t]
        cdef array.array result = array.clone(template, num_rows, 0)
        cdef unsigned long long sz = result.itemsize * num_rows
        cdef char * b = self.read_bytes_c(sz)
        memcpy(result.data.as_voidptr, b, sz)
        if must_swap:
            result.byteswap()
        return result

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_bytes_col(self, unsigned long long sz, unsigned long long num_rows) -> Iterable[Any]:
        cdef object column = PyTuple_New(num_rows)
        cdef char * b = self.read_bytes_c(sz * num_rows)
        for x in range(num_rows):
            v = PyBytes_FromStringAndSize(b, sz)
            b += sz
            PyTuple_SET_ITEM(column, x, v)
            Py_INCREF(v)
        return column

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_fixed_str_col(self, unsigned long long sz, unsigned long long num_rows,
                           encoding:str ='utf8') -> Iterable[str]:
        cdef object column = PyTuple_New(num_rows)
        cdef char * enc
        cdef char * b = self.read_bytes_c(sz * num_rows)
        cdef object v
        pyenc = encoding.encode()
        enc = pyenc
        for x in range(num_rows):
            try:
                v = PyUnicode_Decode(b, sz, enc, errors).rstrip("\x00")
            except UnicodeDecodeError:
                v = PyBytes_FromStringAndSize(b, sz).hex()
            PyTuple_SET_ITEM(column, x, v)
            Py_INCREF(v)
            b += sz
        return column

    def close(self):
        if self.source:
            self.source.close()
            self.source = None

    @property
    def exception_tag(self):
        return self._exception_tag

    @property
    def last_message(self):
        if self.last_message_data is not None:
            return self.last_message_data
        if self.current_chunk is not None:
            return self.current_chunk
        return b""

    def __dealloc__(self):
        self.close()
        PyBuffer_Release(&self.buff_source)
        PyMem_Free(self.slice)



import logging
from typing import Union

from clickhouse_connect.datatypes import registry
from clickhouse_connect.driver.common import write_leb128
from clickhouse_connect.driver.exceptions import StreamCompleteException, StreamFailureError
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.npquery import NumpyResult
from clickhouse_connect.driver.query import QueryResult, QueryContext
from clickhouse_connect.driver.types import ByteSource
from clickhouse_connect.driver.compression import get_compressor

_EMPTY_CTX = QueryContext()

logger = logging.getLogger(__name__)


class NativeTransform:
    # pylint: disable=too-many-locals, too-many-statements, too-many-branches
    @staticmethod
    def parse_response(source: ByteSource, context: QueryContext = _EMPTY_CTX) -> Union[NumpyResult, QueryResult]:
        names = []
        col_types = []
        block_num = 0
        renamer = context.column_renamer

        def get_block():
            nonlocal block_num
            result_block = []
            try:
                try:
                    if context.block_info:
                        source.read_bytes(8)
                    num_cols = source.read_leb128()
                except StreamCompleteException:
                    if source.last_message:
                        error_msg = None
                        exception_tag = getattr(source, "exception_tag", None)
                        if exception_tag:
                            error_msg = extract_exception_with_tag(source.last_message, exception_tag)
                        if error_msg:
                            raise StreamFailureError(error_msg) from None
                    return None
                num_rows = source.read_leb128()
                for col_num in range(num_cols):
                    orig_name = source.read_leb128_str()
                    type_name = source.read_leb128_str()
                    if block_num == 0:
                        disp_name = renamer(orig_name) if renamer is not None else orig_name
                        names.append(disp_name)
                        col_type = registry.get_from_name(type_name)
                        col_types.append(col_type)
                    else:
                        col_type = col_types[col_num]
                    if num_rows == 0:
                        result_block.append(tuple())
                    else:
                        context.start_column(orig_name)
                        column = col_type.read_column(source, num_rows, context)
                        result_block.append(column)
            except Exception as ex:
                source.close()
                if isinstance(ex, StreamCompleteException):
                    # We ran out of data before it was expected, this could be ClickHouse reporting an error
                    # in the response
                    if source.last_message:
                        error_msg = None
                        exception_tag = getattr(source, "exception_tag", None)
                        if exception_tag:
                            error_msg = extract_exception_with_tag(source.last_message, exception_tag)
                        if not error_msg:
                            error_msg = extract_error_message(source.last_message)
                        raise StreamFailureError(error_msg) from None
                raise
            block_num += 1
            return result_block

        first_block = get_block()
        if first_block is None:
            return NumpyResult() if context.use_numpy else QueryResult([])

        def gen():
            yield first_block
            while True:
                next_block = get_block()
                if next_block is None:
                    return
                yield next_block

        if context.use_numpy:
            res_types = [col.dtype if hasattr(col, 'dtype') else 'O' for col in first_block]
            return NumpyResult(gen(), tuple(names), tuple(col_types), res_types, source)
        return QueryResult(None, gen(), tuple(names), tuple(col_types), context.column_oriented, source)

    @staticmethod
    def build_insert(context: InsertContext):
        compressor = get_compressor(context.compression)

        def chunk_gen():
            for block in context.next_block():
                output = bytearray()
                output += block.prefix
                write_leb128(block.column_count, output)
                write_leb128(block.row_count, output)
                for col_name, col_type, data in zip(block.column_names, block.column_types, block.column_data):
                    col_enc = col_name.encode()
                    write_leb128(len(col_enc), output)
                    output += col_enc
                    col_enc = col_type.insert_name.encode()
                    write_leb128(len(col_enc), output)
                    output += col_enc
                    context.start_column(col_name)
                    try:
                        col_type.write_column(data, output, context)
                    except Exception as ex:  # pylint: disable=broad-except
                        # This is hideous, but some low level serializations can fail while streaming
                        # the insert if the user has included bad data in the column.  We need to ensure that the
                        # insert fails (using garbage data) to avoid a partial insert, and use the context to
                        # propagate the correct exception to the user
                        logger.error('Error serializing column `%s` into data type `%s`',
                                     col_name, col_type.name, exc_info=True)
                        context.insert_exception = ex
                        yield 'INTERNAL EXCEPTION WHILE SERIALIZING'.encode()
                        return
                yield compressor.compress_block(output)
            footer = compressor.flush()
            if footer:
                yield footer

        return chunk_gen()


# pylint: disable=too-many-return-statements,too-many-branches
def extract_exception_with_tag(message: bytes, exception_tag: str) -> Union[str, None]:
    """Extract exception message from the new format with exception tag. Server v25.11+.

    Format: __exception__<TAG>\\r\\n<error message>\\r\\n<message_length> <TAG>__exception__\\r\\n
    """
    if not exception_tag:
        return None

    marker = b"__exception__"
    marker_pos = message.find(marker)
    if marker_pos == -1:
        return None

    pos = marker_pos + len(marker)
    while pos < len(message) and message[pos : pos + 1] in (b"\r", b"\n"):
        pos += 1

    tag_end = message.find(b"\r", pos)
    if tag_end == -1:
        tag_end = message.find(b"\n", pos)
    if tag_end == -1:
        return None

    found_tag = message[pos:tag_end].decode("ascii", errors="ignore").strip()
    if found_tag != exception_tag:
        return None

    pos = tag_end
    while pos < len(message) and message[pos : pos + 1] in (b"\r", b"\n"):
        pos += 1

    # Find the footer pattern: <message_length> <TAG>\r\n__exception__
    footer_pattern = f" {exception_tag}".encode()
    footer_pos = message.rfind(footer_pattern)
    if footer_pos == -1 or footer_pos < pos:
        return None

    suffix = message[footer_pos + len(footer_pattern) :]
    if b"__exception__" not in suffix:
        return None

    search_start = max(pos, footer_pos - 100)  # Search last 100 bytes for the newline
    last_newline = message.rfind(b"\n", search_start, footer_pos)
    if last_newline != -1:
        error_end = last_newline
        if error_end > 0 and message[error_end - 1 : error_end] == b"\r":
            error_end -= 1
    else:
        error_end = footer_pos

    error_message = message[pos:error_end]

    try:
        return error_message.decode("utf-8", errors="replace").strip()
    except Exception:  # pylint: disable=broad-except
        return error_message.decode("latin-1", errors="replace").strip()


def extract_error_message(message: bytes) -> str:
    if len(message) > 1024:
        message = message[-1024:]
    error_start = message.find('Code: '.encode())
    if error_start != -1:
        message = message[error_start:]
    try:
        message_str = message.decode()
    except UnicodeError:
        message_str = f'unrecognized data found in stream: `{message.hex()[128:]}`'
    return message_str



import struct
from typing import Sequence, Optional

import array
from datetime import datetime, date

import cython

from .buffer cimport ResponseBuffer
from cpython cimport Py_INCREF, Py_DECREF
from cpython.buffer cimport PyBUF_READ
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM
from cpython.bytearray cimport PyByteArray_GET_SIZE, PyByteArray_Resize
from cpython.memoryview cimport PyMemoryView_FromMemory
from cython.view cimport array as cvarray
from ipaddress import IPv4Address
from uuid import UUID, SafeUUID
from libc.string cimport memcpy
from datetime import tzinfo

from clickhouse_connect.driver import tzutil
from clickhouse_connect.driver.errors import NONE_IN_NULLABLE_COLUMN

@cython.boundscheck(False)
@cython.wraparound(False)
def pivot(data: Sequence, unsigned long long start, unsigned long long end):
    return tuple(zip(*data[start:end]))


@cython.wraparound(False)
@cython.boundscheck(False)
def read_ipv4_col(ResponseBuffer buffer, unsigned long long num_rows):
    cdef unsigned long long x = 0
    cdef char* loc = buffer.read_bytes_c(4 * num_rows)
    cdef object column = PyTuple_New(num_rows), v
    ip_new = IPv4Address.__new__
    while x < num_rows:
        v = ip_new(IPv4Address)
        v._ip = (<unsigned int*>loc)[0]
        PyTuple_SET_ITEM(column, x, v)
        Py_INCREF(v)
        loc += 4
        x += 1
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
def read_datetime_col(ResponseBuffer buffer, unsigned long long num_rows, tzinfo: tzinfo):
    cdef unsigned long long x = 0
    cdef char * loc = buffer.read_bytes_c(4 * num_rows)
    cdef object column = PyTuple_New(num_rows), v
    if tzinfo is None:
        fts = tzutil.utcfromtimestamp
        while x < num_rows:
            v = fts((<unsigned int*>loc)[0])
            PyTuple_SET_ITEM(column, x, v)
            Py_INCREF(v)
            loc += 4
            x += 1
    else:
        fts = datetime.fromtimestamp
        while x < num_rows:
            v = fts((<unsigned int*>loc)[0], tzinfo)
            PyTuple_SET_ITEM(column, x, v)
            Py_INCREF(v)
            loc += 4
            x += 1
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
def read_date_col(ResponseBuffer buffer, unsigned long long num_rows):
    cdef unsigned long long x = 0
    cdef char * loc = buffer.read_bytes_c(2 * num_rows)
    cdef object column = PyTuple_New(num_rows), v
    while x < num_rows:
        v = epoch_days_to_date((<unsigned short*>loc)[0])
        PyTuple_SET_ITEM(column, x, v)
        Py_INCREF(v)
        loc += 2
        x += 1
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
def read_date32_col(ResponseBuffer buffer, unsigned long long num_rows):
    cdef unsigned long long x = 0
    cdef char * loc = buffer.read_bytes_c(4 * num_rows)
    cdef object column = PyTuple_New(num_rows), v
    while x < num_rows:
        v = epoch_days_to_date((<int*>loc)[0])
        PyTuple_SET_ITEM(column, x, v)
        Py_INCREF(v)
        loc += 4
        x += 1
    return column


cdef unsigned short* MONTH_DAYS = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365]
cdef unsigned short* MONTH_DAYS_LEAP = [0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366]

# Constants used in epoch_days_to_date
# 47482 -- Jan 1, 2100 -- Because all years 1970-2099 divisible by 4 are leap years, some extra division can be avoided
# 134774 -- Number of days between Jan 1 1601 and Jan 1 1970.  Adding this starts all calculations at 1601-01-01
# 1461 -- Number of days in a 4-year cycle (365 * 4) + 1 leap day
# 36524 -- Number of days in a 100-year cycle.  25 4-year cycles - 1 leap day for the year 100
# 146097 -- Number of days in a 400-year cycle.  4 100 year cycles + 1 leap day for the year 400

# Year and offset with in the year are determined by factoring out the largest "known" year blocks in
# descending order (400/100/4/1 years).  Month is then (over) estimated in the "day" arrays (days / 32) and
# adjusted down if too large (logic originally in the Python standard library)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef inline object epoch_days_to_date(int days):
    cdef int years, month, year, cycles400, cycles100, cycles, rem
    cdef unsigned short prev
    cdef unsigned short* m_list
    if 0 <= days < 47482:
        cycles = (days + 365) // 1461
        rem = (days + 365) - cycles * 1461
        years = rem // 365
        rem -= years * 365
        year = (cycles << 2) + years + 1969
        if years == 4:
            return date(year - 1, 12, 31)
        if years == 3:
            m_list = MONTH_DAYS_LEAP
        else:
            m_list = MONTH_DAYS
    else:
        cycles400 = (days + 134774) // 146097
        rem = days + 134774 - (cycles400 * 146097)
        cycles100 = rem // 36524
        rem -= cycles100 * 36524
        cycles = rem // 1461
        rem -= cycles * 1461
        years = rem // 365
        rem -= years * 365
        year = (cycles << 2) + cycles400 * 400 + cycles100 * 100  + years + 1601
        if years == 4 or cycles100 == 4:
            return date(year - 1, 12, 31)
        if years == 3 and (year == 2000 or year % 100 != 0):
            m_list = MONTH_DAYS_LEAP
        else:
            m_list = MONTH_DAYS
    month = (rem + 24) >> 5
    prev = m_list[month]
    while rem < prev:
        month -= 1
        prev = m_list[month]
    return date(year, month + 1, rem + 1 - prev)


@cython.boundscheck(False)
@cython.wraparound(False)
def read_uuid_col(ResponseBuffer buffer, unsigned long long num_rows):
    cdef unsigned long long x = 0
    cdef char * loc = buffer.read_bytes_c(16 * num_rows)
    cdef char[16] temp
    cdef object column = PyTuple_New(num_rows), v
    new_uuid = UUID.__new__
    unsafe = SafeUUID.unsafe
    oset = object.__setattr__
    for x in range(num_rows):
        memcpy (<void *>temp, <void *>(loc + 8), 8)
        memcpy (<void *>(temp + 8), <void *>loc, 8)
        v = new_uuid(UUID)
        oset(v, 'int', int.from_bytes(temp[:16], 'little'))
        oset(v, 'is_safe', unsafe)
        PyTuple_SET_ITEM(column, x, v)
        Py_INCREF(v)
        loc += 16
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
def read_nullable_array(ResponseBuffer buffer, array_type: str, unsigned long long num_rows, object null_obj):
    if num_rows == 0:
        return []
    cdef unsigned long long x = 0
    cdef size_t item_size = struct.calcsize(array_type)
    cdef cvarray cy_array = cvarray((num_rows,), item_size, array_type, mode='c', allocate_buffer=False)

    # We have to make a copy of the incoming null map because the next
    # "read_byes_c" call could invalidate our pointer by replacing the underlying buffer
    cdef char * null_map = <char *>PyMem_Malloc(<size_t>num_rows)
    memcpy(<void *>null_map, <void *>buffer.read_bytes_c(num_rows), num_rows)

    cy_array.data = buffer.read_bytes_c(num_rows * item_size)
    cdef object column = list(memoryview(cy_array))
    for x in range(num_rows):
        if null_map[x] != 0:
            column[x] = null_obj
    PyMem_Free(<void *>null_map)
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
def build_nullable_column(source: Sequence, char * null_map, object null_obj):
    cdef unsigned long long num_rows = len(source), x
    cdef object column = [], v
    cdef object app = column.append
    for x in range(num_rows):
        if null_map[x] == 0:
            v = source[x]
        else:
            v = null_obj
        app(v)
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
def build_lc_nullable_column(index: Sequence, keys: array.array, object null_obj):
    cdef unsigned long long num_rows = len(keys), x, y
    cdef object column = [], v
    cdef object app = column.append
    for x in range(num_rows):
        y = keys[x]
        if y == 0:
            v = null_obj
        else:
            v = index[y]
        app(v)
    return column


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline extend_byte_array(target: bytearray, int start, object source, Py_ssize_t sz):
    PyByteArray_Resize(target, start + sz)
    target[start:start + sz] = source[0:sz]


@cython.boundscheck(False)
@cython.wraparound(False)
def write_str_col(column: Sequence, nullable: bool, encoding: Optional[str], dest: bytearray) -> int:
    cdef unsigned long long buff_size = len(column) << 5
    cdef unsigned long long buff_loc = 0, sz = 0, dsz = 0
    cdef unsigned long long array_size = PyByteArray_GET_SIZE(dest)
    cdef char * temp_buff = <char *>PyMem_Malloc(<size_t>buff_size)
    cdef object mv = PyMemoryView_FromMemory(temp_buff, buff_size, PyBUF_READ)
    cdef object encoded
    cdef char b
    cdef char * data
    try:
        for x in column:
            if not x:
                if not nullable and x is None:
                    return NONE_IN_NULLABLE_COLUMN
                temp_buff[buff_loc] = 0
                buff_loc += 1
                if buff_loc == buff_size:
                    extend_byte_array(dest, array_size, mv, buff_loc)
                    array_size += buff_loc
                    buff_loc = 0
            else:
                if not encoding:
                    data = x
                    dsz = len(x)
                else:
                    encoded = x.encode(encoding)
                    dsz = len(encoded)
                    data = encoded
                sz = dsz
                while True:
                    b = sz & 0x7f
                    sz >>= 7
                    if sz != 0:
                        b |= 0x80
                    temp_buff[buff_loc] = b
                    buff_loc += 1
                    if buff_loc == buff_size:
                        extend_byte_array(dest, array_size, mv, buff_loc)
                        array_size += buff_loc
                        buff_loc = 0
                    if sz == 0:
                        break
                if dsz + buff_loc >= buff_size:
                    if buff_loc > 0:  # Write what we have so far
                        extend_byte_array(dest, array_size, mv, buff_loc)
                        array_size += buff_loc
                        buff_loc = 0
                    if (dsz << 4) > buff_size:  # resize our buffer for very large strings
                        PyMem_Free(<void *> temp_buff)
                        mv.release()
                        buff_size = dsz << 6
                        temp_buff = <char *> PyMem_Malloc(<size_t> buff_size)
                        mv = PyMemoryView_FromMemory(temp_buff, buff_size, PyBUF_READ)
                memcpy(temp_buff + buff_loc, data, dsz)
                buff_loc += dsz
        if buff_loc > 0:
            extend_byte_array(dest, array_size, mv, buff_loc)
    finally:
        mv.release()
        PyMem_Free(<void *>temp_buff)
    return 0



from typing import Optional, Union

from sqlalchemy import Table
from sqlalchemy.sql.selectable import FromClause, Select

from clickhouse_connect.driver.binding import quote_identifier

# Dialect name used for non-rendering statement hints that only serve to
# differentiate cache keys when FINAL/SAMPLE modifiers are applied.
_CH_MODIFIER_DIALECT = "_ch_modifier"


def full_table(table_name: str, schema: Optional[str] = None) -> str:
    if table_name.startswith('(') or '.' in table_name or not schema:
        return quote_identifier(table_name)
    return f'{quote_identifier(schema)}.{quote_identifier(table_name)}'


def format_table(table: Table):
    return full_table(table.name, table.schema)


def _resolve_target(select_stmt: Select, table: Optional[FromClause], method_name: str) -> FromClause:
    """Resolve the target FROM clause for ClickHouse modifiers (FINAL/SAMPLE)."""
    if not isinstance(select_stmt, Select):
        raise TypeError(f"{method_name}() expects a SQLAlchemy Select instance")

    target = table
    if target is None:
        froms = select_stmt.get_final_froms()
        if not froms:
            raise ValueError(f"{method_name}() requires a table to apply the {method_name.upper()} modifier.")
        if len(froms) > 1:
            raise ValueError(
                f"{method_name}() is ambiguous for statements with multiple FROM clauses. "
                "Specify the table explicitly."
            )
        target = froms[0]

    if not isinstance(target, FromClause):
        raise TypeError("table must be a SQLAlchemy FromClause when provided")

    return target


def _target_cache_key(target: FromClause) -> str:
    """Stable string identifying a FROM target for cache key differentiation."""
    if hasattr(target, "fullname"):
        return target.fullname
    return target.name


# pylint: disable=protected-access
def final(select_stmt: Select, table: Optional[FromClause] = None) -> Select:
    """Apply the ClickHouse FINAL modifier to a select statement.

    FINAL forces ClickHouse to merge data parts before returning results,
    guaranteeing fully collapsed rows for ReplacingMergeTree, CollapsingMergeTree,
    and similar engines.

    Args:
        select_stmt: The SELECT statement to modify.
        table: The target table to apply FINAL to. Required when the query
            joins multiple tables, optional when there is a single FROM target.
    """
    target = _resolve_target(select_stmt, table, "final")
    ch_final = getattr(select_stmt, "_ch_final", set())

    if target in ch_final:
        return select_stmt

    # with_statement_hint creates a generative copy and adds a non-rendering
    # hint that participates in the statement cache key.
    hint_key = _target_cache_key(target)
    new_stmt = select_stmt.with_statement_hint(f"FINAL:{hint_key}", dialect_name=_CH_MODIFIER_DIALECT)
    new_stmt._ch_final = ch_final | {target}
    return new_stmt


def _select_final(self: Select, table: Optional[FromClause] = None) -> Select:
    """
    Select.final() convenience wrapper around the module-level final() helper.
    """
    return final(self, table=table)


def sample(select_stmt: Select, sample_value: Union[str, int, float], table: Optional[FromClause] = None) -> Select:
    """Apply the ClickHouse SAMPLE modifier to a select statement.

    Args:
        select_stmt: The SELECT statement to modify.
        sample_value: The sample expression. Can be a float between 0 and 1
            for a fractional sample (e.g. 0.1 for 10%), an integer for an
            approximate row count, or a string for SAMPLE expressions like
            '1/10 OFFSET 1/2'.
        table: The target table to sample. Required when the query joins
            multiple tables, optional when there is a single FROM target.
    """
    target = _resolve_target(select_stmt, table, "sample")

    hint_key = _target_cache_key(target)
    new_stmt = select_stmt.with_statement_hint(
        f"SAMPLE:{hint_key}:{sample_value}", dialect_name=_CH_MODIFIER_DIALECT
    )
    ch_sample = dict(getattr(select_stmt, "_ch_sample", {}))
    ch_sample[target] = sample_value
    new_stmt._ch_sample = ch_sample
    return new_stmt


def _select_sample(self: Select, sample_value: Union[str, int, float], table: Optional[FromClause] = None) -> Select:
    """
    Select.sample() convenience wrapper around the module-level sample() helper.
    """
    return sample(self, sample_value=sample_value, table=table)


# Monkey-patch the select class to add final and sample methods
Select.sample = _select_sample
Select.final = _select_final



import socket
from ipaddress import ip_address, IPv4Address, IPv6Address
from typing import Union, MutableSequence, Sequence, Any

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.common import write_array, int_size, first_value
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.query import QueryContext
from clickhouse_connect.driver.types import ByteSource
from clickhouse_connect.driver.ctypes import data_conv

IPV4_V6_MASK = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff'
V6_NULL = bytes(b'\x00' * 16)


# pylint: disable=protected-access
class IPv4(ClickHouseType):
    _array_type = 'L' if int_size == 2 else 'I'
    valid_formats = 'string', 'native', 'int'
    python_type = IPv4Address
    byte_size = 4

    def _read_column_binary(self, source: ByteSource, num_rows: int, ctx: QueryContext, _read_state: Any):
        if self.read_format(ctx) == 'int':
            return source.read_array(self._array_type, num_rows)
        if self.read_format(ctx) == 'string':
            column = source.read_array(self._array_type, num_rows)
            return [socket.inet_ntoa(x.to_bytes(4, 'big')) for x in column]
        return data_conv.read_ipv4_col(source, num_rows)

    def _write_column_binary(self, column: Union[Sequence, MutableSequence], dest: bytearray, ctx: InsertContext):
        first = first_value(column, self.nullable)
        if isinstance(first, str):
            fixed = 24, 16, 8, 0
            # pylint: disable=consider-using-generator
            column = [(sum([int(b) << fixed[ix] for ix, b in enumerate(x.split('.'))])) if x else 0 for x in column]
        else:
            if self.nullable:
                column = [x._ip if x else 0 for x in column]
            else:
                column = [x._ip for x in column]
        write_array(self._array_type, column, dest, ctx.column_name)

    def _active_null(self, ctx: QueryContext):
        fmt = self.read_format(ctx)
        if ctx.use_none:
            return None
        if fmt == 'string':
            return '0.0.0.0'
        if fmt == 'int':
            return 0
        return None


# pylint: disable=protected-access
class IPv6(ClickHouseType):
    valid_formats = 'string', 'native'
    python_type = IPv6Address
    byte_size = 16

    def _read_column_binary(self, source: ByteSource, num_rows: int, ctx: QueryContext, _read_state: Any):
        if self.read_format(ctx) == 'string':
            return self._read_binary_str(source, num_rows)
        return self._read_binary_ip(source, num_rows)

    @staticmethod
    def _read_binary_ip(source: ByteSource, num_rows: int) -> list[IPv6Address]:
        """Read IPv6 addresses in native format, always returning IPv6Address objects."""
        fast_ip_v6 = IPv6Address.__new__
        with_scope_id = '_scope_id' in IPv6Address.__slots__
        new_col = []
        app = new_col.append
        ifb = int.from_bytes
        for _ in range(num_rows):
            int_value = ifb(source.read_bytes(16), 'big')
            ipv6 = fast_ip_v6(IPv6Address)
            ipv6._ip = int_value
            if with_scope_id:
                ipv6._scope_id = None
            app(ipv6)
        return new_col

    @staticmethod
    def _read_binary_str(source: ByteSource, num_rows: int) -> list[str]:
        """Read IPv6 addresses in string format, always returning IPv6Address strings."""
        new_col = []
        app = new_col.append
        tov6 = socket.inet_ntop
        af6 = socket.AF_INET6
        for _ in range(num_rows):
            x = source.read_bytes(16)
            # Always use IPv6 string representation, even for IPv4-mapped addresses
            app(tov6(af6, x))
        return new_col

    def _write_column_binary(
        self,
        column: Union[Sequence, MutableSequence],
        dest: bytearray,
        ctx: InsertContext,
    ):
        """Write IPv6 addresses, promoting IPv4 addresses to IPv4-mapped IPv6 addresses."""
        for value in column:
            if value is None:
                dest += V6_NULL
                continue

            try:
                addr = ip_address(value)
            except ValueError as e:
                raise ValueError(
                    f"Failed to parse '{value}' as a valid IP address for column '{ctx.column_name}'"
                ) from e

            # Now handle parsed object
            if isinstance(addr, IPv6Address):
                dest += addr.packed
            elif isinstance(addr, IPv4Address):
                # We have an IPv4, but the column is IPv6 so convert to IPv4-mapped.
                dest += IPV4_V6_MASK + addr.packed

    def _active_null(self, ctx):
        if ctx.use_none:
            return None
        return '::' if self.read_format(ctx) == 'string' else V6_NULL



import os
from datetime import datetime, tzinfo
from typing import Optional, Tuple, Union

import pytz

tzlocal = None
try:
    import tzlocal  # Maybe we can use the tzlocal module to get a safe timezone
except ImportError:
    pass

# Set the local timezone for DateTime conversions.  Note in most cases we want to use either UTC or the server
# timezone, but if someone insists on using the local timezone we will try to convert.  The problem is we
# never have anything but an epoch timestamp returned from ClickHouse, so attempts to convert times when the
# local timezone is "DST" aware (like 'CEST' vs 'CET') will be wrong approximately half the time
local_tz: pytz.timezone
local_tz_dst_safe: bool = False

# Timezone names that are equivalent to UTC
UTC_EQUIVALENTS = ('UTC', 'Etc/UTC', 'GMT', 'Universal', 'GMT-0', 'Zulu', 'Greenwich', 'UCT')


def normalize_timezone(timezone: pytz.timezone) -> Tuple[pytz.timezone, bool]:
    if timezone.tzname(None) in UTC_EQUIVALENTS:
        return pytz.UTC, True

    if timezone.tzname(None) in pytz.common_timezones:
        return timezone, True

    if tzlocal is not None:  # Maybe we can use the tzlocal module to get a safe timezone
        local_name = tzlocal.get_localzone_name()
        if local_name in pytz.common_timezones:
            return pytz.timezone(local_name), True

    return timezone, False


def is_utc_timezone(tz: Optional[Union[tzinfo, str]]) -> bool:
    """Check if timezone is UTC or an equivalent (Etc/UTC, GMT, etc.).

    This handles the issue where pytz.timezone('Etc/UTC') != pytz.UTC despite
    being semantically equivalent. Also accepts timezone name strings.
    """
    if tz is None:
        return False
    if isinstance(tz, str):
        return tz in UTC_EQUIVALENTS
    if tz == pytz.UTC:
        return True
    return tz.tzname(None) in UTC_EQUIVALENTS


def utcfromtimestamp(ts: float) -> datetime:
    return datetime.fromtimestamp(ts, tz=pytz.UTC).replace(tzinfo=None)

try:
    local_tz = pytz.timezone(os.environ.get('TZ', ''))
except pytz.UnknownTimeZoneError:
    local_tz = datetime.now().astimezone().tzinfo

local_tz, local_tz_dst_safe = normalize_timezone(local_tz)



import logging
from math import ceil, nan
from struct import pack, unpack
from typing import Any, Sequence

from clickhouse_connect.datatypes.base import ClickHouseType, TypeDef
from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.driver.ctypes import data_conv
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver import options
from clickhouse_connect.driver.query import QueryContext
from clickhouse_connect.driver.types import ByteSource

logger = logging.getLogger(__name__)


class QBit(ClickHouseType):
    """
    QBit type - represents bit-transposed vectors for efficient vector search operations.

    Syntax: QBit(element_type, dimension)
    - element_type: BFloat16, Float32, or Float64
    - dimension: Number of elements per vector

    Over the Native protocol, ClickHouse transmits QBit columns as bit-transposed Tuples.

    Requires:
        - SET allow_experimental_qbit_type = 1
        - Server version >=25.10
    """

    __slots__ = (
        "element_type",
        "dimension",
        "_bits_per_element",
        "_bytes_per_fixedstring",
        "_tuple_type",
    )

    python_type = list
    _BIT_SHIFTS = [1 << i for i in range(8)]
    _ELEMENT_BITS = {"BFloat16": 16, "Float32": 32, "Float64": 64}
    _numpy_warned = False

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        if not QBit._numpy_warned and options.np is None:
            QBit._numpy_warned = True
            logger.info("NumPy not detected. Install NumPy to see an order of magnitude performance gain with QBit columns.")

        self.element_type = type_def.values[0]
        if self.element_type not in self._ELEMENT_BITS:
            raise ValueError(f"Unsupported QBit element type '{self.element_type}'. Supported types: BFloat16, Float32, Float64.")

        self.dimension = type_def.values[1]
        if self.dimension <= 0:
            raise ValueError(f"QBit dimension must be greater than 0. Got: {self.dimension}.")

        self._name_suffix = f"({self.element_type}, {self.dimension})"
        self._bits_per_element = self._ELEMENT_BITS.get(self.element_type, 32)
        self._bytes_per_fixedstring = ceil(self.dimension / 8)

        # Create the underlying Tuple type for bit-transposed representation
        # E.g., for Float32 with dim=8: Tuple(FixedString(1), FixedString(1), ... x32)
        fixedstring_type = f"FixedString({self._bytes_per_fixedstring})"
        tuple_types = ", ".join([fixedstring_type] * self._bits_per_element)
        tuple_type_name = f"Tuple({tuple_types})"
        self._tuple_type = get_from_name(tuple_type_name)
        self.byte_size = self._bits_per_element * self._bytes_per_fixedstring

    def read_column_prefix(self, source: ByteSource, ctx: QueryContext):
        return self._tuple_type.read_column_prefix(source, ctx)

    def read_column_data(self, source: ByteSource, num_rows: int, ctx: QueryContext, read_state: Any) -> Sequence:
        """Read bit-transposed Tuple data and convert to flat float vectors."""
        if num_rows == 0:
            return []

        null_map = None
        if self.nullable:
            null_map = source.read_bytes(num_rows)

        tuple_data = self._tuple_type.read_column_data(source, num_rows, ctx, read_state)
        vectors = [self._untranspose_row(t) for t in tuple_data]
        if self.nullable:
            return data_conv.build_nullable_column(vectors, null_map, self._active_null(ctx))
        return vectors

    def write_column_prefix(self, dest: bytearray):
        self._tuple_type.write_column_prefix(dest)

    def write_column_data(self, column: Sequence, dest: bytearray, ctx: InsertContext):
        """Convert flat float vectors to bit-transposed Tuple data and write."""
        if len(column) == 0:
            return

        if self.nullable:
            dest += bytes([1 if x is None else 0 for x in column])

        null_tuple = tuple(b"\x00" * self._bytes_per_fixedstring for _ in range(self._bits_per_element))
        tuple_column = [null_tuple if row is None else self._transpose_row(row) for row in column]

        self._tuple_type.write_column_data(tuple_column, dest, ctx)

    def _active_null(self, ctx: QueryContext):
        """Return context-appropriate null value for nullable QBit columns."""
        if ctx.use_none:
            return None
        if ctx.use_extended_dtypes:
            return nan
        return None

    def _values_to_words(self, values: list[float]) -> Sequence[int]:
        """Convert float values to integer words using batch struct processing."""
        count = len(values)

        if self.element_type == "BFloat16":
            # BFloat16 is the top 16 bits of a Float32 (truncate mantissa)
            raw_ints = unpack(f"<{count}I", pack(f"<{count}f", *values))
            return [(x >> 16) & 0xFFFF for x in raw_ints]

        fmt_char = "I" if self.element_type == "Float32" else "Q"
        float_char = "f" if self.element_type == "Float32" else "d"

        return unpack(f"<{count}{fmt_char}", pack(f"<{count}{float_char}", *values))

    def _words_to_values(self, words: list[int]) -> list[float]:
        """Convert integer words to float values using batch unpacking."""
        count = len(words)

        if self.element_type == "BFloat16":
            # Pad BFloat16 words with zeros to reconstruct valid Float32s
            shifted_words = [(w & 0xFFFF) << 16 for w in words]
            return list(unpack(f"<{count}f", pack(f"<{count}I", *shifted_words)))

        if self.element_type == "Float32":
            return list(unpack(f"<{count}f", pack(f"<{count}I", *words)))

        # Float64
        return list(unpack(f"<{count}d", pack(f"<{count}Q", *words)))

    def _untranspose_row(self, bit_planes: tuple):
        """Convert bit-transposed tuple to flat float vector."""
        if options.np is not None:
            return self._untranspose_row_numpy(bit_planes)

        words = [0] * self.dimension
        bit_shifts = self._BIT_SHIFTS
        dim = self.dimension

        # Iterate Planes (MSB -> LSB)
        for bit_idx, bit_plane_bytes in enumerate(bit_planes):
            bit_pos = self._bits_per_element - 1 - bit_idx
            mask = 1 << bit_pos

            # Iterate Bytes in Plane
            for byte_idx, byte_val in enumerate(bit_plane_bytes):
                # if byte is 0, skip processing 8 bits
                if byte_val == 0:
                    continue

                base_elem_idx = byte_idx << 3  # Each byte encodes 8 elements

                # Extract set bits from this byte
                for bit_in_byte in range(8):
                    if byte_val & bit_shifts[bit_in_byte]:
                        elem_idx = base_elem_idx + bit_in_byte
                        if elem_idx < dim:
                            words[elem_idx] |= mask  # Accumulate bit at position bit_pos

        return self._words_to_values(words)

    def _untranspose_row_numpy(self, bit_planes: tuple) -> list[float]:
        """Vectorized numpy operations version of _untranspose_row"""
        # 1. Convert tuple of bytes to a single uint8 array
        total_bytes = b"".join(bit_planes)
        planes_uint8 = options.np.frombuffer(total_bytes, dtype=options.np.uint8)
        planes_uint8 = planes_uint8.reshape(self._bits_per_element, -1)

        # 2. Unpack bits to get the boolean/integer matrix
        bits_matrix: "np.ndarray" = options.np.unpackbits(planes_uint8, axis=1, bitorder="little")

        # 3. Trim padding if necessary
        if bits_matrix.shape[1] != self.dimension:  # pylint: disable=no-member
            bits_matrix = bits_matrix[:, : self.dimension]  # pylint: disable=invalid-sequence-index

        # 4. Reconstruct the integer words
        if self.element_type == "Float64":
            int_dtype = options.np.uint64
            final_dtype = options.np.float64
        else:
            # Float32 and BFloat16 use 32-bit containers
            int_dtype = options.np.uint32
            final_dtype = options.np.float32

        # Accumulate bits into integers
        words = options.np.zeros(self.dimension, dtype=int_dtype)

        for i in range(self._bits_per_element):
            # MSB is at index 0
            shift = self._bits_per_element - 1 - i

            # If the bit row is 1, add 2^shift to the word
            # Cast bits to the target int type before shifting to avoid overflow
            words |= bits_matrix[i].astype(int_dtype) << shift

        # 5. Interpret as Floats
        if self.element_type == "BFloat16":
            # Shift back up to the top 16 bits of a Float32
            # Cast to uint32 first to ensure safe shifting
            words = words.astype(options.np.uint32) << 16
            return words.view(options.np.float32).tolist()

        return words.view(final_dtype).tolist()

    def _transpose_row(self, values: list[float]) -> tuple:
        """Convert flat float vector to bit-transposed tuple."""
        if len(values) != self.dimension:
            raise ValueError(f"Vector dimension mismatch: expected {self.dimension}, got {len(values)}")

        # If numpy is available, use the fast path
        if options.np is not None:
            if isinstance(values, options.np.ndarray):
                return self._transpose_row_numpy(values)

            # If numpy is available but user supplied python list, convert to np array anyway for
            #  huge performance gains.
            dtype = options.np.float64 if self.element_type == "Float64" else options.np.float32
            return self._transpose_row_numpy(options.np.array(values, dtype=dtype))

        words = self._values_to_words(values)
        bit_planes = []
        bit_shifts = self._BIT_SHIFTS
        bytes_per_fs = self._bytes_per_fixedstring

        for bit_idx in range(self._bits_per_element):
            bit_pos = self._bits_per_element - 1 - bit_idx
            mask = 1 << bit_pos
            plane = bytearray(bytes_per_fs)

            for elem_idx, word in enumerate(words):
                if word & mask:
                    plane[elem_idx >> 3] |= bit_shifts[elem_idx & 7]

            bit_planes.append(bytes(plane))

        return tuple(bit_planes)

    def _transpose_row_numpy(self, vector: "np.ndarray") -> tuple:
        """Fast path for numpy arrays using vectorized operations."""
        # Cast to int view
        if self.element_type == "BFloat16":
            # Numpy doesn't have bfloat16. Input is Float32 so just
            #  discard the bottom 16 bits.
            v_float = vector.astype(options.np.float32, copy=False)
            # View as uint32, shift right 16, cast to uint16
            v_int = (v_float.view(options.np.uint32) >> 16).astype(options.np.uint16)

        elif self.element_type == "Float32":
            # Ensure it is 32-bit float first (handles float64->32 downcast safely)
            v_float = vector.astype(options.np.float32, copy=False)
            v_int = v_float.view(options.np.uint32)

        else:  # Float64
            v_float = vector.astype(options.np.float64, copy=False)
            v_int = v_float.view(options.np.uint64)

        bits = self._bits_per_element
        masks = (1 << options.np.arange(bits - 1, -1, -1, dtype=v_int.dtype)).reshape(-1, 1)

        # Extract bits: (Bits, Dim)
        # v_int broadcasted to (1, Dim)
        bits_extracted = (v_int & masks) != 0

        packed = options.np.packbits(bits_extracted.view(options.np.uint8), axis=1, bitorder="little")

        return tuple(row.tobytes() for row in packed)



import clickhouse_connect


def test_bare_install():
    """Bare install test to validate the package works with only core dependencies"""
    client = clickhouse_connect.get_client()

    ver = client.command("SELECT version()")
    assert isinstance(ver, str) and len(ver) > 0, f"unexpected version: {ver}"

    result = client.query("SELECT 1 AS x, 2 AS y")
    assert result.result_rows == [(1, 2)], f"unexpected result: {result.result_rows}"

    client.command("DROP TABLE IF EXISTS _bare_install_test")
    client.command("CREATE TABLE _bare_install_test (id UInt32, val String) ENGINE MergeTree ORDER BY id")
    client.insert("_bare_install_test", [[1, "a"], [2, "b"]], column_names=["id", "val"])
    res = client.query("SELECT * FROM _bare_install_test ORDER BY id")
    assert res.result_rows == [(1, "a"), (2, "b")], f"unexpected: {res.result_rows}"
    client.command("DROP TABLE _bare_install_test")



import pytz

import array
from datetime import date, datetime, tzinfo, timedelta, time

from typing import Union, Sequence, MutableSequence, Any, NamedTuple, Optional
from abc import abstractmethod
import re

from clickhouse_connect.datatypes.base import TypeDef, ClickHouseType
from clickhouse_connect.common import get_setting
from clickhouse_connect.driver import tzutil
from clickhouse_connect.driver.common import write_array, np_date_types, int_size, first_value
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver import ctypes as driver_ctypes
from clickhouse_connect.driver.ctypes import data_conv
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.query import QueryContext
from clickhouse_connect.driver.types import ByteSource
from clickhouse_connect.driver import options

epoch_start_date = date(1970, 1, 1)
epoch_start_datetime = datetime(1970, 1, 1)


class Date(ClickHouseType):
    _array_type = 'H'
    np_type = 'datetime64[D]'
    nano_divisor = 86400 * 1000000000
    valid_formats = 'native', 'int'
    python_type = date
    byte_size = 2

    @property
    def pandas_dtype(self):
        if options.IS_PANDAS_2 and get_setting("preserve_pandas_datetime_resolution"):
            return "datetime64[s]"
        return f"datetime64[{self.pd_datetime_res}]"

    def _read_column_binary(self, source: ByteSource, num_rows: int, ctx: QueryContext, _read_state:Any):
        if self.read_format(ctx) == 'int':
            return source.read_array(self._array_type, num_rows)
        if ctx.use_numpy:
            return driver_ctypes.numpy_conv.read_numpy_array(source, '<u2', num_rows).astype(self.np_type)
        return data_conv.read_date_col(source, num_rows)

    def _write_column_binary(self, column: Union[Sequence, MutableSequence], dest: bytearray, ctx: InsertContext):
        first = first_value(column, self.nullable)
        if isinstance(first, int) or self.write_format(ctx) == 'int':
            if self.nullable:
                column = [x if x else 0 for x in column]
        else:
            if isinstance(first, datetime):
                esd = epoch_start_datetime
            else:
                esd = epoch_start_date
            if self.nullable:
                column = [0 if x is None else (x - esd).days for x in column]
            else:
                column = [(x - esd).days for x in column]
        write_array(self._array_type, column, dest, ctx.column_name)

    def _active_null(self, ctx: QueryContext):
        fmt = self.read_format(ctx)
        if ctx.use_extended_dtypes:
            return options.pd.NA if fmt == 'int' else options.pd.NaT
        if ctx.use_none:
            return None
        if fmt == 'int':
            return 0
        if ctx.use_numpy:
            return options.np.datetime64(0, self.pd_datetime_res)
        return epoch_start_date

    # pylint: disable=too-many-return-statements
    def _finalize_column(self, column: Sequence, ctx: QueryContext) -> Sequence:
        if self.read_format(ctx) == 'int':
            return column

        if ctx.use_numpy and self.nullable and not ctx.use_none:
            return options.np.array(column, dtype=self.np_type)

        if ctx.use_extended_dtypes:
            if isinstance(column, options.np.ndarray) and options.np.issubdtype(
                column.dtype, options.np.datetime64
            ):
                return column.astype(self.pandas_dtype)

            if isinstance(column, options.pd.DatetimeIndex):
                if column.tz is None:
                    return column.astype(self.pandas_dtype)

                naive = column.tz_convert("UTC").tz_localize(None).astype(self.pandas_dtype)
                return naive.tz_localize("UTC").tz_convert(column.tz)

            if self.nullable and isinstance(column, list):
                return options.np.array([None if options.pd.isna(s) else s for s in column]).astype(
                    self.pandas_dtype
                )

            return options.pd.to_datetime(column, errors="coerce").to_numpy(
                dtype=self.pandas_dtype, copy=False
            )

        return column


class Date32(Date):
    byte_size = 4
    _array_type = 'l' if int_size == 2 else 'i'

    def _read_column_binary(self, source: ByteSource, num_rows: int, ctx: QueryContext, _read_state: Any):
        if ctx.use_numpy:
            return driver_ctypes.numpy_conv.read_numpy_array(source, '<i4', num_rows).astype(self.np_type)
        if self.read_format(ctx) == 'int':
            return source.read_array(self._array_type, num_rows)
        return data_conv.read_date32_col(source, num_rows)


class DateTimeBase(ClickHouseType, registered=False):
    __slots__ = ('tzinfo',)
    valid_formats = 'native', 'int'
    python_type = datetime

    @property
    def pandas_dtype(self):
        """Sets dtype for pandas datetime objects"""
        if options.IS_PANDAS_2 and get_setting("preserve_pandas_datetime_resolution"):
            return "datetime64[s]"
        return f"datetime64[{self.pd_datetime_res}]"

    def _active_null(self, ctx: QueryContext):
        fmt = self.read_format(ctx)
        if ctx.use_extended_dtypes:
            return options.pd.NA if fmt == 'int' else options.pd.NaT
        if ctx.use_none:
            return None
        if self.read_format(ctx) == 'int':
            return 0
        if ctx.use_numpy:
            return options.np.datetime64(0, self.pd_datetime_res)
        return epoch_start_datetime

    def _finalize_column(self, column: Sequence, ctx: QueryContext) -> Sequence:
        """Ensure every datetime-like column is at nanosecond resolution, preserving any tz."""
        if ctx.use_extended_dtypes:
            if isinstance(column, options.np.ndarray) and options.np.issubdtype(
                column.dtype, options.np.datetime64
            ):
                return column.astype(self.pandas_dtype)

            if isinstance(column, options.pd.DatetimeIndex) or (
                isinstance(column, list)
                and hasattr(next((s for s in column if not options.pd.isna(s)), None), "tz")
            ):
                if isinstance(column, list):
                    column = options.pd.DatetimeIndex(column)

                if column.tz is None:
                    result = column.astype(self.pandas_dtype)
                    return options.pd.array(result) if self.nullable else result

                naive_ns = column.tz_convert("UTC").tz_localize(None).astype(self.pandas_dtype)
                tz_aware_result = naive_ns.tz_localize("UTC").tz_convert(column.tz)
                return (
                    options.pd.array(tz_aware_result) if self.nullable else tz_aware_result
                )

            if self.nullable:
                return options.pd.array(
                    [None if options.pd.isna(s) else s for s in column], dtype=self.pandas_dtype
                )
        return column


class DateTime(DateTimeBase):
    _array_type = 'L' if int_size == 2 else 'I'
    np_type = 'datetime64[s]'
    nano_divisor = 1000000000
    byte_size = 4

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str
        if len(type_def.values) > 0:
            self.tzinfo = pytz.timezone(type_def.values[0][1:-1])
        else:
            self.tzinfo = None

    def _read_column_binary(self, source: ByteSource, num_rows: int, ctx: QueryContext, _read_state: Any) -> Sequence:
        if self.read_format(ctx) == 'int':
            return source.read_array(self._array_type, num_rows)
        active_tz = ctx.active_tz(self.tzinfo)
        if ctx.use_numpy:
            np_array = driver_ctypes.numpy_conv.read_numpy_array(source, '<u4', num_rows).astype(self.np_type)
            if ctx.as_pandas and active_tz:
                return options.pd.DatetimeIndex(np_array, tz='UTC').tz_convert(active_tz)
            return np_array
        return data_conv.read_datetime_col(source, num_rows, active_tz)

    def _write_column_binary(self, column: Union[Sequence, MutableSequence], dest: bytearray, ctx: InsertContext):
        first = first_value(column, self.nullable)
        if isinstance(first, int) or self.write_format(ctx) == 'int':
            if self.nullable:
                column = [x if x else 0 for x in column]
        else:
            if self.nullable:
                column = [int(x.timestamp()) if x else 0 for x in column]
            else:
                column = [int(x.timestamp()) for x in column]
        write_array(self._array_type, column, dest, ctx.column_name)


class DateTime64(DateTimeBase):
    __slots__ = 'scale', 'prec', 'unit'
    byte_size = 8

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str
        self.scale = type_def.values[0]
        self.prec = 10 ** self.scale
        self.unit = np_date_types.get(self.scale)
        if len(type_def.values) > 1:
            self.tzinfo = pytz.timezone(type_def.values[1][1:-1])
        else:
            self.tzinfo = None

    @property
    def pandas_dtype(self):
        """Sets dtype for pandas datetime objects"""
        if options.IS_PANDAS_2 and get_setting("preserve_pandas_datetime_resolution"):
            return f"datetime64{self.unit}"
        return f"datetime64[{self.pd_datetime_res}]"

    @property
    def np_type(self):
        if self.unit:
            return f'datetime64{self.unit}'
        raise ProgrammingError(f'Cannot use {self.name} as a numpy or Pandas datatype. Only milliseconds(3), ' +
                               'microseconds(6), or nanoseconds(9) are supported for numpy based queries.')

    @property
    def nano_divisor(self):
        return 1000000000 // self.prec

    def _read_column_binary(self, source: ByteSource, num_rows: int, ctx: QueryContext, _read_state: Any) -> Sequence:
        if self.read_format(ctx) == 'int':
            return source.read_array('q', num_rows)
        active_tz = ctx.active_tz(self.tzinfo)
        if ctx.use_numpy:
            np_array = driver_ctypes.numpy_conv.read_numpy_array(source, self.np_type, num_rows)
            if ctx.as_pandas and active_tz:
                return options.pd.DatetimeIndex(np_array, tz='UTC').tz_convert(active_tz)
            return np_array
        column = source.read_array('q', num_rows)
        if active_tz:
            return self._read_binary_tz(column, active_tz)
        return self._read_binary_naive(column)

    def _read_binary_tz(self, column: Sequence, tz_info: tzinfo):
        new_col = []
        app = new_col.append
        dt_from = datetime.fromtimestamp
        prec = self.prec
        for ticks in column:
            seconds = ticks // prec
            dt_sec = dt_from(seconds, tz_info)
            app(dt_sec.replace(microsecond=((ticks - seconds * prec) * 1000000) // prec))
        return new_col

    def _read_binary_naive(self, column: Sequence):
        new_col = []
        app = new_col.append
        dt_from = tzutil.utcfromtimestamp
        prec = self.prec
        for ticks in column:
            seconds = ticks // prec
            dt_sec = dt_from(seconds)
            app(dt_sec.replace(microsecond=((ticks - seconds * prec) * 1000000) // prec))
        return new_col

    def _write_column_binary(self, column: Union[Sequence, MutableSequence], dest: bytearray, ctx: InsertContext):
        first = first_value(column, self.nullable)
        if isinstance(first, int) or self.write_format(ctx) == 'int':
            if self.nullable:
                column = [x if x else 0 for x in column]
        elif isinstance(first, str):
            original_column = column
            column = []

            for x in original_column:
                if not x and self.nullable:
                    v = 0
                else:
                    dt = datetime.fromisoformat(x)
                    v = ((int(dt.timestamp()) * 1000000 + dt.microsecond) * self.prec) // 1000000

                column.append(v)
        else:
            prec = self.prec
            if self.nullable:
                column = [((int(x.timestamp()) * 1000000 + x.microsecond) * prec) // 1000000 if x else 0
                          for x in column]
            else:
                column = [((int(x.timestamp()) * 1000000 + x.microsecond) * prec) // 1000000 for x in column]
        write_array('q', column, dest, ctx.column_name)


class _HMSParts(NamedTuple):
    """Internal structure for parsed HMS time components."""

    hours: int
    minutes: int
    seconds: int
    frac: Optional[str]
    is_negative: bool


class TimeBase(ClickHouseType, registered=False):
    """
    Abstract base for ClickHouse Time and Time64 types.

    Subclasses must define:
      - _array_type: Array type specifier (e.g. 'i' or 'q')
      - byte_size: Size in bytes for binary representation
      - np_type: NumPy array type (e.g. 'timedelta64[s]' or 'timedelta64[ns]')

    And implement these abstract methods:
      - _string_to_ticks(self, str) -> int
      - _timedelta_to_ticks(self, timedelta) -> int
      - _ticks_to_timedelta(self, int) -> timedelta
      - _ticks_to_string(self, int) -> str
      - max_ticks and min_ticks properties
    """

    _HMS_RE = re.compile(
        r"""^\s*
        (?P<sign>-?)
        (?P<hours>\d+):
        (?P<minutes>\d+):
        (?P<seconds>\d+)
        (?:\.(?P<frac>\d+))?
        \s*$""",
        re.VERBOSE,
    )

    MAX_TIME_SECONDS = 999 * 3600 + 59 * 60 + 59  # 999:59:59
    MIN_TIME_SECONDS = -MAX_TIME_SECONDS  # -999:59:59
    _MICROS_PER_SECOND = 1_000_000
    _NANOS_PER_SECOND = 1_000_000_000
    _SECONDS_PER_DAY = 86_400

    _array_type: str
    byte_size: int
    np_type: str
    valid_formats = ("native", "string", "int", "time")
    python_type = timedelta

    def _read_column_binary(
        self,
        source: ByteSource,
        num_rows: int,
        ctx: QueryContext,
        _read_state: Any,
    ) -> Sequence:
        """Read binary column data and convert to requested format."""
        ticks = source.read_array(self._array_type, num_rows)
        fmt = self.read_format(ctx)

        if ctx.use_numpy:
            return options.np.array(
                [self._ticks_to_np_timedelta(t) for t in ticks], dtype=self.np_type
            )

        if fmt == "int":
            return ticks

        if fmt == "string":
            return [self._ticks_to_string(t) for t in ticks]

        if fmt == "time":
            return [self._ticks_to_time(t) for t in ticks]

        return [self._ticks_to_timedelta(t) for t in ticks]

    def _write_column_binary(
        self,
        column: Sequence,
        dest: bytearray,
        ctx: InsertContext,
    ):
        """Write column data in binary format."""
        ticks = self._to_ticks_array(column)
        write_array(self._array_type, ticks, dest, ctx.column_name)

    def _parse_core(self, time_str: str) -> _HMSParts:
        """Parse an hhh:mm:ss[.fff] time literal."""
        match = self._HMS_RE.match(time_str)
        if not match:
            raise ValueError(f"Invalid time literal {time_str}")

        hours = int(match["hours"])
        minutes = int(match["minutes"])
        seconds = int(match["seconds"])

        if hours > 999:
            raise ValueError(
                f"Hours out of range; cannot exceed 999: got {hours} in '{time_str}'"
            )
        if not 0 <= minutes < 60:
            raise ValueError(
                f"Minutes out of range; must be 0-59: got {minutes} in '{time_str}'"
            )
        if not 0 <= seconds < 60:
            raise ValueError(
                f"Seconds out of range; must be 0-59: got {seconds} in '{time_str}'"
            )

        return _HMSParts(
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            frac=match["frac"],
            is_negative=bool(match["sign"]),
        )

    def _to_ticks_array(self, column: Sequence) -> Sequence[int]:
        """Convert column data to internal tick representation."""
        first = first_value(column, self.nullable)
        expected_type = type(first) if first is not None else None

        if expected_type is None:
            if self.nullable:
                return [0] * len(column)
            return []

        converter_map = {
            timedelta: self._timedelta_to_ticks,
            time: self._time_to_ticks,
            float: self._numerical_to_ticks,
            int: self._numerical_to_ticks,
            str: self._string_to_ticks,
        }
        if options.np is not None:
            converter_map[options.np.timedelta64] = self._timedelta_to_ticks
            converter_map[options.np.int64] = self._numerical_to_ticks
        converter = converter_map.get(expected_type, None)

        if converter is None:
            raise TypeError(
                f"Unsupported column type '{expected_type.__name__}' for {self.__class__.__name__}. "
                "Expected 'int', 'str', 'time', or 'timedelta'."
            )

        if self.nullable:
            return [converter(x) if x is not None else 0 for x in column]

        return [converter(x) for x in column]

    def _validate_standard_range(self, ticks: int, original: Any) -> None:
        """Validate that ticks is within valid ClickHouse range."""
        if not self.min_ticks <= ticks <= self.max_ticks:
            raise ValueError(f"{original} out of range for {self.__class__.__name__}")

    def _validate_time_obj_range(self, ticks: int) -> None:
        """Ensure ticks can form a valid datetime.time object."""
        if not self.min_time_ticks <= ticks <= self.max_time_ticks:
            raise ValueError(
                f"Ticks value {ticks} is outside valid range for datetime.time object."
            )

    def _numerical_to_ticks(self, value: Union[int, float, "np.int64"]) -> int:
        """Convert numerical value to ticks, with range validation."""
        value = int(value)
        self._validate_standard_range(value, value)
        return value

    def _active_null(self, ctx: QueryContext):
        """Return appropriate null value based on context."""
        fmt = self.read_format(ctx)
        if ctx.use_extended_dtypes:
            return options.pd.NA if fmt == "int" else options.pd.NaT
        if ctx.use_none:
            return None
        if fmt == "int":
            return 0
        if fmt == "string":
            return "00:00:00"
        if ctx.use_numpy:
            return options.np.timedelta64("NaT")

        return timedelta(0)

    @property
    def pandas_dtype(self):
        """Sets dtype for pandas datetime objects"""
        if options.IS_PANDAS_2 and get_setting("preserve_pandas_datetime_resolution"):
            return "timedelta64[s]"
        return f"timedelta64[{self.pd_datetime_res}]"

    def _finalize_column(self, column: Sequence, ctx: QueryContext) -> Sequence:
        """Finalize column data based on context requirements."""
        if ctx.use_extended_dtypes:
            if isinstance(column, options.np.ndarray) and options.np.issubdtype(
                column.dtype, options.np.timedelta64
            ):
                return column.astype(self.pandas_dtype)

            if isinstance(column, options.pd.TimedeltaIndex):
                return column.astype(self.pandas_dtype)

            if self.nullable:
                return options.np.array([None if options.pd.isna(s) else s for s in column]).astype(
                    self.pandas_dtype
                )
        return column

    def _build_lc_column(self, index: Sequence, keys: array.array, ctx: QueryContext):
        """Build low-cardinality column from index and keys."""
        if ctx.use_numpy:
            return options.np.array([index[k] for k in keys], dtype=self.np_type)

        return super()._build_lc_column(index, keys, ctx)

    @abstractmethod
    def _string_to_ticks(self, time_str: str) -> int:
        """Parse a string into integer ticks."""
        raise NotImplementedError

    @abstractmethod
    def _timedelta_to_ticks(self, td: Union[timedelta, "np.timedelta64"]) -> int:
        """Convert a timedelta into integer ticks."""
        raise NotImplementedError

    @abstractmethod
    def _ticks_to_time(self, ticks: int) -> time:
        """Convert integer ticks into a time."""
        raise NotImplementedError

    @abstractmethod
    def _time_to_ticks(self, t: time) -> int:
        """Convert a time into integer ticks."""
        raise NotImplementedError

    @abstractmethod
    def _ticks_to_timedelta(self, ticks: int) -> timedelta:
        """Convert integer ticks into a timedelta."""
        raise NotImplementedError

    @abstractmethod
    def _ticks_to_np_timedelta(self, ticks: int) -> timedelta:
        """Convert integer ticks into an np.timedelta."""
        raise NotImplementedError

    @abstractmethod
    def _ticks_to_string(self, ticks: int) -> str:
        """Format integer ticks as a string."""
        raise NotImplementedError

    @property
    def min_time_ticks(self) -> int:
        """Minimum tick value representable by datetime.time type."""
        return 0

    @property
    @abstractmethod
    def max_time_ticks(self) -> int:
        """Maximum tick value representable by datetime.time type."""
        raise NotImplementedError

    @property
    @abstractmethod
    def max_ticks(self) -> int:
        """Maximum tick value representable by this type."""
        raise NotImplementedError

    @property
    @abstractmethod
    def min_ticks(self) -> int:
        """Minimum tick value representable by this type."""
        raise NotImplementedError


class Time(TimeBase):
    """ClickHouse Time type with second precision."""

    _array_type = "i"
    byte_size = 4
    np_type = "timedelta64[s]"

    @property
    def max_ticks(self) -> int:
        return self.MAX_TIME_SECONDS

    @property
    def min_ticks(self) -> int:
        return self.MIN_TIME_SECONDS

    @property
    def max_time_ticks(self) -> int:
        return self._SECONDS_PER_DAY - 1

    def _string_to_ticks(self, time_str: str) -> int:
        """Parse string format 'HHH:MM:SS[.fff]' to ticks (seconds), flooring fractional seconds."""
        parts = self._parse_core(time_str)
        ticks = parts.hours * 3600 + parts.minutes * 60 + parts.seconds

        if parts.is_negative:
            ticks = -ticks
        self._validate_standard_range(ticks, time_str)

        return ticks

    def _ticks_to_string(self, ticks: int) -> str:
        """Format ticks (seconds) as 'HHH:MM:SS' string."""
        sign = "-" if ticks < 0 else ""
        t = abs(ticks)
        h, rem = divmod(t, 3600)
        m, s = divmod(rem, 60)

        return f"{sign}{h:03d}:{m:02d}:{s:02d}"

    def _timedelta_to_ticks(self, td: Union[timedelta, "np.timedelta64"]) -> int:
        """Convert timedelta to ticks (seconds), flooring fractional seconds."""
        if isinstance(td, timedelta):
            total = int(td.total_seconds())
        else:
            total = td.astype("timedelta64[s]").astype(int)
        self._validate_standard_range(total, td)

        return total

    def _ticks_to_timedelta(self, ticks: int) -> timedelta:
        """Convert ticks (seconds) to timedelta."""
        return timedelta(seconds=ticks)

    def _ticks_to_np_timedelta(self, ticks: int) -> timedelta:
        """Convert ticks (seconds) to np.timedelta."""
        return options.np.timedelta64(ticks, "s")

    def _time_to_ticks(self, t: time) -> int:
        """Converts time to ticks (seconds), flooring fraction seconds."""
        return t.hour * 3600 + t.minute * 60 + t.second

    def _ticks_to_time(self, ticks: int) -> time:
        """Converts ticks (seconds) to time."""
        self._validate_time_obj_range(ticks)
        h, rem = divmod(ticks, 3600)
        m, s = divmod(rem, 60)

        return time(hour=h, minute=m, second=s)


class Time64(TimeBase):
    """ClickHouse Time64 type with configurable sub-second precision."""

    __slots__ = ("scale", "precision", "unit")
    _array_type = "q"
    byte_size = 8

    def __init__(self, type_def):
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str
        self.scale = type_def.values[0]
        if self.scale not in (3, 6, 9):
            raise ProgrammingError(
                f"Unsupported Time64 scale {self.scale}; "
                "only 3, 6, or 9 are allowed for NumPy."
            )
        self.precision = 10**self.scale
        self.unit = np_date_types.get(self.scale)

    @property
    def pandas_dtype(self):
        """Sets dtype for pandas datetime objects"""
        if options.IS_PANDAS_2 and get_setting("preserve_pandas_datetime_resolution"):
            return f"timedelta64{self.unit}"
        return f"timedelta64[{self.pd_datetime_res}]"

    @property
    def max_time_ticks(self) -> int:
        return self._SECONDS_PER_DAY * self.precision - 1

    @property
    def np_type(self) -> str:
        return f"timedelta64{self.unit}"

    @property
    def max_ticks(self) -> int:
        return self.MAX_TIME_SECONDS * self.precision + (self.precision - 1)

    @property
    def min_ticks(self) -> int:
        return -self.max_ticks

    def _string_to_ticks(self, time_str: str) -> int:
        """Parse string format 'HHH:MM:SS[.fff]' to ticks with sub-second precision."""
        parts = self._parse_core(time_str)
        frac_ticks = int((parts.frac or "").ljust(self.scale, "0")[: self.scale])
        ticks = (
            parts.hours * 3600 + parts.minutes * 60 + parts.seconds
        ) * self.precision + frac_ticks
        if parts.is_negative:
            ticks = -ticks
        self._validate_standard_range(ticks, time_str)

        return ticks

    def _ticks_to_string(self, ticks: int) -> str:
        """Format ticks as 'HHH:MM:SS[.fff]' string with sub-second precision."""
        sign = "-" if ticks < 0 else ""
        t = abs(ticks)
        sec_part, frac_part = divmod(t, self.precision)
        h, rem = divmod(sec_part, 3600)
        m, s = divmod(rem, 60)
        frac_str = f".{frac_part:0{self.scale}d}" if self.scale else ""

        return f"{sign}{h:03d}:{m:02d}:{s:02d}{frac_str}"

    def _timedelta_to_ticks(self, td: Union[timedelta, "np.timedelta64"]) -> int:
        """Convert timedelta to ticks with sub-second precision."""
        if isinstance(td, timedelta):
            total_us = (
                int(td.total_seconds()) * self._MICROS_PER_SECOND + td.microseconds
            )
            ticks = (total_us * self.precision) // self._MICROS_PER_SECOND
        else:
            ticks = td.astype("timedelta64[s]").astype(int)
        self._validate_standard_range(ticks, td)

        return ticks

    def _ticks_to_timedelta(self, ticks: int) -> timedelta:
        """Convert ticks to timedelta with microsecond precision."""
        neg = ticks < 0
        t = abs(ticks)
        sec_part = t // self.precision
        frac_part = t - sec_part * self.precision
        micros = (frac_part * self._MICROS_PER_SECOND) // self.precision
        td = timedelta(seconds=sec_part, microseconds=micros)

        return -td if neg else td

    def _ticks_to_np_timedelta(self, ticks: int) -> "np.timedelta64":
        """Convert ticks to numpy timedelta64 with nanosecond precision."""
        res_map = {3: "ms", 6: "us", 9: "ns"}

        return options.np.timedelta64(ticks, res_map.get(self.scale))

    def _time_to_ticks(self, t: time) -> int:
        """Convert time to ticks with sub-second precision."""
        total_us = (
            t.hour * 3600 + t.minute * 60 + t.second
        ) * self._MICROS_PER_SECOND + t.microsecond
        ticks = (total_us * self.precision) // self._MICROS_PER_SECOND
        self._validate_time_obj_range(ticks)

        return ticks

    def _ticks_to_time(self, ticks: int) -> time:
        """Convert ticks to time with microsecond precision."""
        self._validate_time_obj_range(ticks)
        sec_part, frac_part = divmod(ticks, self.precision)
        h, rem = divmod(sec_part, 3600)
        m, s = divmod(rem, 60)
        micros = (frac_part * self._MICROS_PER_SECOND) // self.precision

        return time(hour=h, minute=m, second=s, microsecond=micros)



#!/usr/bin/env python3

# This script is used for validating installed entrypoints.  Note that it fails on Python 3.7
import sys

from importlib.metadata import PackageNotFoundError, distribution

EXPECTED_EPS = {'sqlalchemy.dialects:clickhousedb',
                'sqlalchemy.dialects:clickhousedb.connect'}


def validate_entrypoints():
    expected_eps = EXPECTED_EPS.copy()
    try:
        dist = distribution('clickhouse-connect')
    except PackageNotFoundError:
        print ('\nClickHouse Connect package not found in this Python installation')
        return -1
    print()
    for entry_point in dist.entry_points:
        name = f'{entry_point.group}:{entry_point.name}'
        print(f'    {name}={entry_point.value}')
        try:
            expected_eps.remove(name)
        except KeyError:
            print (f'\nUnexpected entry point {name} found')
            return -1
    if expected_eps:
        print()
        for name in expected_eps:
            print (f'Did not find expected ep {name}')
        return -1
    print ('\nEntrypoints correctly installed')
    return 0


if __name__ == '__main__':
    sys.exit(validate_entrypoints())



import logging
import re

from typing import Optional, Sequence, List, Dict

from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.driver.common import unescape_identifier
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver import Client
from clickhouse_connect.driver.parser import parse_callable
from clickhouse_connect.driver.query import remove_sql_comments

logger = logging.getLogger(__name__)

insert_re = re.compile(r'^\s*INSERT\s+INTO\s+(.*$)', re.IGNORECASE)
str_type = get_from_name('String')
int_type = get_from_name('Int32')


# pylint: disable=too-many-instance-attributes
class Cursor:
    """
    See :ref:`https://peps.python.org/pep-0249/`
    """

    def __init__(self, client: Client):
        self.client = client
        self.arraysize = 1
        self.data: Optional[Sequence] = None
        self.names = []
        self.types = []
        self._rowcount = 0
        self._summary: List[Dict[str, str]] = []
        self._ix = 0

    def check_valid(self):
        if self.data is None:
            raise ProgrammingError('Cursor is not valid')

    @property
    def description(self):
        return [(n, t, None, None, None, None, True) for n, t in zip(self.names, self.types)]

    @property
    def rowcount(self):
        return self._rowcount

    @property
    def summary(self) -> List[Dict[str, str]]:
        return self._summary

    def close(self):
        self.data = None

    def execute(self, operation: str, parameters=None):
        if not parameters and isinstance(operation, str):
            # Per PEP 249 pyformat paramstyle, callers (e.g. SQLAlchemy) escape
            # literal percent signs as %% in operation strings.  When there are
            # parameters, Python's % operator in finalize_query handles the
            # unescaping automatically.  When there are no parameters,
            # finalize_query short-circuits, so we must unescape here.
            operation = operation.replace('%%', '%')
        query_result = self.client.query(operation, parameters)
        self.data = query_result.result_set
        self._rowcount = len(self.data)
        self._summary.append(query_result.summary)

        # Need to reset cursor _ix after performing an execute
        self._ix = 0

        if query_result.column_names:
            self.names = query_result.column_names
            self.types = [x.name for x in query_result.column_types]
        elif self.data:
            self.names = [f'col_{x}' for x in range(len(self.data[0]))]
            self.types = [x.__class__ for x in self.data[0]]
        else:
            stripped = operation.strip().rstrip(";").strip()
            if stripped.upper().startswith(("SELECT", "WITH")):
                meta_result = self.client.query(f"SELECT * FROM ({stripped}) LIMIT 0", parameters)
                if meta_result.column_names:
                    self.names = meta_result.column_names
                    self.types = [x.name for x in meta_result.column_types]

    def _try_bulk_insert(self, operation: str, data):
        match = insert_re.match(remove_sql_comments(operation))
        if not match:
            return False
        temp = match.group(1)
        table_end = min(temp.find(' '), temp.find('('))
        table = temp[:table_end].strip()
        temp = temp[table_end:].strip()
        if temp[0] == '(':
            _, op_columns, temp = parse_callable(temp)
        else:
            op_columns = None
        if 'VALUES' not in temp.upper():
            return False
        col_names = list(data[0].keys())
        if op_columns and {unescape_identifier(x) for x in op_columns} != set(col_names):
            return False  # Data sent in doesn't match the columns in the insert statement
        data_values = [list(row.values()) for row in data]
        self.client.insert(table, data_values, col_names)
        self.data = []
        return True

    def executemany(self, operation, parameters):
        if not parameters or self._try_bulk_insert(operation, parameters):
            return
        self.data = []
        try:
            for param_row in parameters:
                query_result = self.client.query(operation, param_row)
                self.data.extend(query_result.result_set)
                if self.names or self.types:
                    if query_result.column_names != self.names:
                        logger.warning('Inconsistent column names %s : %s for operation %s in cursor executemany',
                                       self.names, query_result.column_names, operation)
                else:
                    self.names = query_result.column_names
                    self.types = query_result.column_types
                self._summary.append(query_result.summary)
        except TypeError as ex:
            raise ProgrammingError(f'Invalid parameters {parameters} passed to cursor executemany') from ex
        self._rowcount = len(self.data)

        # Need to reset cursor _ix after performing an execute
        self._ix = 0

    def fetchall(self):
        self.check_valid()
        ret = self.data[self._ix:]
        self._ix = self._rowcount
        return ret

    def fetchone(self):
        self.check_valid()
        if self._ix >= self._rowcount:
            return None
        val = self.data[self._ix]
        self._ix += 1
        return val

    def fetchmany(self, size: int = -1):
        self.check_valid()

        if size < 0:
            # Fetch all remaining rows
            size = self._rowcount - self._ix
        elif size == 0:
            # Return empty list for size=0
            return []

        end = min(self._ix + size, self._rowcount)
        ret = self.data[self._ix: end]
        self._ix = end
        return ret

    def nextset(self):
        raise NotImplementedError

    def callproc(self, *args, **kwargs):
        raise NotImplementedError
